# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions for working with tarballs."""

from __future__ import annotations

import re
import typing as t

from antsibull_core.subprocess_util import async_log_run

if t.TYPE_CHECKING:
    from _typeshed import StrPath

#: Regex to find toplevel directories in tar output
TOPLEVEL_RE: re.Pattern = re.compile("^[^/]+/$")


class InvalidTarball(Exception):
    """Raised when a requested version does not exist."""


async def unpack_tarball(tarname: StrPath, destdir: StrPath) -> str:
    """
    Unpack a tarball.

    :arg tarname: The tarball to unpack.
    :arg destdir: The destination to unpack into.
    :returns: Toplevel of the unpacked directory structure.  This will be
        a subdirectory of `destdir`.
    """
    tarname_str = str(tarname)
    manifest = await async_log_run(["tar", "-xzvf", tarname_str, f"-C{destdir}"])
    toplevel_dirs = [
        filename
        for filename in manifest.stdout.splitlines()
        if TOPLEVEL_RE.match(filename)
    ]

    if len(toplevel_dirs) != 1:
        raise InvalidTarball(
            f"The tarball {tarname} had more than a single toplevel dir"
        )

    expected_dirname = tarname_str[: -len(".tar.gz")]
    if toplevel_dirs[0] != expected_dirname:
        raise InvalidTarball(f"The directory in {tarname} was not {expected_dirname}")

    return toplevel_dirs[0]


async def pack_tarball(tarname: StrPath, directory: str) -> str:
    raise NotImplementedError("pack_tarball has not yet been implemented")



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2019, Toshio Kuratomi
"""
Setup an application context to save global data which is set on program start.

There is some application data which is set on application startup and then never modified.
These values are *almost* constants and should be eligible to be global variables.  However,
global variables make testing hard (as the test has to monkeypatch constants) and the use of
globals prevents re-using the functionality as a library.

Enter contexts.  Contexts can be used to set global values.  But the context can be swapped out if
the program needs to use a different set of globals and some location for some reason.  This file
contains the context framework itself.

For Python3.6 compatibility, this file needs to be loaded early, before any event loops are created.
This is due to limitations in the backport of the contextvars library to Python3.6.  For code which
targets Python3.7 and above, there is no such limitation.

.. warning:: This is not a stable interface.

    The API has quite a few rough edges that need to be ironed out before this is finished.  Some of
    this code and data will be moved into an antibull.context module which can deal with the generic
    side of things while this module will only contain the things that are particular to specfic
    applications.

Setup
=====

Importing antsibull_core.app_context will setup a default context with default values for the
library to use.  The application should initialize a new context with user overriding values by
calling :func:`antsibull_core.app_context.create_contexts` with command line args and
configuration data.  The data from those will be used to initialize a new ``app_ctx`` and new
``lib_ctx``.  The application can then use the context managers to utilize these contexts before
calling any further antsibull code.  An example:

.. code-block:: python

    from antsibull_core import app_context
    from antsibull_core.config import load_config


    def do_something():
        app_ctx = app_context.app_ctx.get()
        core_filename = download_python_package('ansible-core', server_url=app_ctx.pypi_url)
        return core_filename

    def run(args):
        args = parse_args(args)
        cfg = load_config(args.config_file)
        context_data = app_context.create_contexts(args=args, cfg=cfg)
        with app_context.app_and_lib_context(context_data):
            do_something()

Extending AppContext
====================

Since antsibull-core 1.2.0, applications using antsibull-core can use their own extension
of AppContext to better handle command line arguments, or handle additional configuration
values in explicitly specified configuration files.  (The implicitly specified configuration
files, ``/etc/antsibull.cfg`` and ``~/.antsibull.cfg``, cannot have extra keys to prevent
incompatibility with other antsibull-core based applications.)

For this, the application needs to create a derived class of :obj:`AppContext`, and pass
it to :func:`antsibull_core.config.load_config` when loading the configuration.  Please
note that :func:`antsibull_core.app_context.app_ctx.get` always returns the :obj:`AppContext`
view of that extended app context, so applications should create their own ``app_context``
module that provides itself a way to obtain the extended app context.  For example in
antsibull-docs this is done as follows:

.. code-block:: python

    from antsibull_core.app_context import AppContextWrapper
    from antsibull_docs.schemas.app_context import DocsAppContext

    app_ctx: AppContextWrapper[DocsAppContext] = AppContextWrapper()

In antsibull-docs, the extended app context (``DocsAppContext``) is then used as follows:

.. code-block:: python

    from antsibull_core.app_context import app_and_lib_context, create_contexts
    from antsibull_core.config import load_config

    from antsibull_docs import app_context
    from antsibull_docs.schemas.app_context import DocsAppContext


    def do_something():
        app_ctx = app_context.app_ctx.get()
        # collection_url is an option in DocsAppContext
        print(f'collection_url configuration: {app_ctx.collection_url}')

    def run(args):
        args = parse_args(args)
        cfg = load_config(args.config_file, app_context_model=DocsAppContext)
        context_data = create_contexts(args=args, cfg=cfg)
        with app_and_lib_context(context_data):
            do_something()
"""

from __future__ import annotations

import argparse
import contextvars
import typing as t
from collections.abc import Iterable, Mapping
from contextlib import AbstractContextManager, contextmanager

from .schemas.context import AppContext, LibContext
from .vendored.collections import ImmutableDict

AppContextT = t.TypeVar("AppContextT", bound=AppContext)


#: lib_ctx should be restricted to things which do not belong in the API but an application or
#: user might want to tweak.  Global, internal, incidental values are good to store here.  Things
#: that are already settable by the public API are not.  For instance, a function whose primary
#: purpose is to retrieve a file from the internet and return the filename where it was
#: downloaded to might need a number of bytes to read at a time so that the entire file contents
#: aren't in memory at one time.  The number of bytes is incidental and an internal
#: implementation detail.  However, it might be something that an end user wants to adjust
#: globally for all functions which need to chunk data.  So this is appropriate to make available
#: for tweaking via a value saved in lib_ctx.  All values in lib_ctx need to have a default value
#: so that code which uses it can fallback to something if the application or user did not
#: specify a value.
lib_ctx: contextvars.ContextVar[LibContext] = contextvars.ContextVar("lib_ctx")

#: Values in app_ctx are things that form defaults in the application.  Even though it may be
#: tempting to use them for library API, they should not be used there.  Instead, these values
#: are things that should be pat of the API calls themselves and explicitly passed from the
#: application to the library code.  If the value is used by multiple calls to the function (or
#: by calls to multiple related functions) it may be convenient to encapsulate that library code
#: into an object which can be initialized with the data.
#:
#: For instance, a function might contact a web service to retrieve information.  The URL of the
#: web service can be passed in via the API for testing against a non-production server.  The
#: user might toggle these via a config file or command line argument.  The app_ctx provides
#: a place for the application to consolidate the information from these different locations into
#: a single place and then consult them globally.  The values should be passed explicitly from
#: the application code to the library code as a function parameter.
#:
#: If the library provides several functions to retrieve different pieces of information from the
#: server, the library can provide a class which takes the server's URL as a parameter and stores
#: as an attribute and the functions can be converted into methods of the object.  Then the
#: application code can initialize the object once and thereafter call the object's methods.
app_ctx: contextvars.ContextVar[AppContext] = contextvars.ContextVar("app_ctx")


class ContextReturn(t.Generic[AppContextT]):
    """
    NamedTuple-like object for the return value of :func:`create_contexts`.

    The :func:`create_contexts` returns quite a bit of information.  This data structure organizes
    the information.

    :ivar app_ctx: Context for vars that are okay to use globally only within application code.
    :ivar lib_ctx: Context for vars which may be used globally within both library and
        application code.
    :ivar args: An :python:obj:`argparse.Namespace` containing command line arguments that were not
        used to construct the contexts
    :ivar cfg: Configuration values which were not used to construct the contexts.

    .. note:: unfortunately generic ``NamedTuple`` objects are not possible, so this is a generic
              class that tries to behave as close as possible to a named tuple. Right now it does
              not support comparisons though, if that is needed please create an issue in the
              antsibull-core repository.
    """

    app_ctx: AppContextT
    lib_ctx: LibContext
    args: argparse.Namespace
    cfg: dict

    def __init__(
        self,
        # pylint: disable-next=redefined-outer-name
        app_ctx: AppContextT,
        # pylint: disable-next=redefined-outer-name
        lib_ctx: LibContext,
        args: argparse.Namespace,
        cfg: dict,
    ):
        self.app_ctx = app_ctx
        self.lib_ctx = lib_ctx
        self.args = args
        self.cfg = cfg

    def __getitem__(self, index: int) -> t.Any:
        if index == 0:
            return self.app_ctx
        if index == 1:
            return self.lib_ctx
        if index == 2:
            return self.args
        if index == 3:
            return self.cfg
        raise IndexError("tuple index out of range")

    def __iter__(self) -> Iterable:
        return (self.app_ctx, self.lib_ctx, self.args, self.cfg).__iter__()


def _extract_context_values(
    known_fields, args: argparse.Namespace | None, cfg: Mapping = ImmutableDict()
) -> dict:
    context_values = {}
    if cfg:
        for value in known_fields:
            try:
                context_values[value] = cfg[value]
            except KeyError:
                pass

    # Args override config
    if args:
        for value in known_fields:
            try:
                context_values[value] = getattr(args, value)
            except AttributeError:
                pass

    return context_values


@t.overload
def create_contexts(
    args: argparse.Namespace | None = None,
    cfg: Mapping = ImmutableDict(),
    use_extra: bool = True,
) -> ContextReturn[AppContext]: ...


@t.overload
def create_contexts(
    args: argparse.Namespace | None = None,
    cfg: Mapping = ImmutableDict(),
    use_extra: bool = True,
    *,
    app_context_model: type[AppContextT],
) -> ContextReturn[AppContextT]: ...


def create_contexts(
    args: argparse.Namespace | None = None,
    cfg: Mapping = ImmutableDict(),
    use_extra: bool = True,
    *,
    app_context_model=AppContext,
) -> ContextReturn:
    """
    Create new contexts appropriate for setting the app and lib context.

    This function takes values from the application arguments and configuration and sets them on
    the context.  It validates, normalizes, and sets defaults for the contexts based on what is
    available in the arguments and configuration.

    :kwarg args: An :python:obj:`argparse.Namespace` holding the program's command line
        arguments.  See the warning below about working with :python:mod:`argpase`.
    :kwarg cfg: A dictionary holding the program's configuration.
    :kwarg use_extra: When True, the default, all extra arguments and config values will be set as
        fields in ``app_ctx.extra``.  When False, the extra arguments and config values will be
        returned as part of the ContextReturn.
    :kwarg app_context_model: The model to use for the app context. Must be derived from
        :obj:`AppContext`. If not provided, will use :obj:`AppContext` itself.
    :returns: A ``ContextReturn`` object.

    .. warning::
        We cannot tell whether a user set a value via the command line if :python:mod:`argparse`
        sets the field to a default value.  That means when you specify the field in the
        :obj:`AppContext` or :obj:`LibContext` models, you must tell :python:mod:`argparse` not to
        set the field to a default like this::

            parser.add_argument('--breadcrumbs', default=argparse.SUPPRESS)

        If the field is only used via the :attr:`AppContext.extra` mechanism (not explictly set),
        then you should ignore this section and use :python:mod:`argparse`'s default mechanism.
    """
    fields_in_lib_ctx = set(LibContext.model_fields)
    fields_in_app_ctx = set(app_context_model.model_fields)
    known_fields = fields_in_app_ctx.union(fields_in_lib_ctx)

    normalized_cfg = dict(cfg)

    lib_values = _extract_context_values(fields_in_lib_ctx, args, normalized_cfg)
    app_values = _extract_context_values(fields_in_app_ctx, args, normalized_cfg)

    #
    # Save the unused values
    #

    unused_cfg = {}
    if normalized_cfg:
        unused_cfg = {k: v for k, v in normalized_cfg.items() if k not in known_fields}

    unused_args = {}
    if args:
        unused_args = {k: v for k, v in vars(args).items() if k not in known_fields}

    # Unused values are saved in app_ctx.extra when use_extra is set
    if use_extra:
        unused_cfg.update(unused_args)
        app_values["extra"] = unused_cfg
        unused_cfg = {}
        unused_args = {}

    unused_args_ns = argparse.Namespace(**unused_args)

    # create new app and lib ctxt from the application's arguments and config.
    local_app_ctx = app_context_model(**app_values)
    local_lib_ctx = LibContext(**lib_values)

    return ContextReturn(
        app_ctx=local_app_ctx,
        lib_ctx=local_lib_ctx,
        args=unused_args_ns,
        cfg=unused_cfg,
    )


def _copy_lib_context() -> LibContext:
    try:
        old_context = lib_ctx.get()
    except LookupError:
        old_context = LibContext()

    # Copy just in case contexts are allowed to be writable in the the future
    return old_context.model_copy()


def _copy_app_context() -> AppContext:
    try:
        old_context = app_ctx.get()
    except LookupError:
        old_context = AppContext()

    # Copy just in case contexts are allowed to be writable in the the future
    return old_context.model_copy()


@contextmanager
def lib_context(
    new_context: LibContext | None = None,
) -> t.Generator[LibContext, None, None]:
    """
    Set up a new lib_context.

    :kwarg new_context: New lib context to setup.  If this is None, the context is set to a copy of
        the old context.
    """
    if new_context is None:
        new_context = _copy_lib_context()

    reset_token = lib_ctx.set(new_context)
    try:
        yield new_context
    finally:
        lib_ctx.reset(reset_token)


@t.overload
def app_context() -> AbstractContextManager[AppContext]: ...


@t.overload
def app_context(new_context: AppContextT) -> AbstractContextManager[AppContextT]: ...


@contextmanager
def app_context(new_context=None):
    """
    Set up a new app_context.

    :kwarg new_context: New app context to setup.
    """
    if new_context is None:
        new_context = _copy_app_context()

    reset_token = app_ctx.set(new_context)
    try:
        yield new_context
    finally:
        app_ctx.reset(reset_token)


@contextmanager
def app_and_lib_context(
    context_data: ContextReturn[AppContextT],
) -> t.Generator[tuple[AppContextT, LibContext], None, None]:
    """
    Set the app and lib context at the same time.

    This is a convenience wrapper around the :func:`app_context` and :func:`lib_context`
    context managers.  It's meant to be used with :func:`create_contexts` like this:

    .. code_block:: python

        context_data = create_contexts(args=args, cfg=cfg)

        with app_and_lib_context(context_data):
            do_something()
    """
    with lib_context(context_data.lib_ctx) as new_lib_ctx:
        with app_context(context_data.app_ctx) as new_app_ctx:
            yield (new_app_ctx, new_lib_ctx)


class AppContextWrapper(t.Generic[AppContextT]):
    def __repr__(self):
        return "<ContextVarWrapper name='app_ctx'>"

    @property
    def name(self):
        return "app_ctx"

    @staticmethod
    def get() -> AppContextT:
        return t.cast(AppContextT, app_ctx.get())


#
# Set initial contexts with default values
#
lib_ctx.set(LibContext())
app_ctx.set(AppContext())



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project
"""I/O helper functions."""

from __future__ import annotations

import typing as t

from antsibull_fileutils.io import copy_file as _copy_file
from antsibull_fileutils.io import read_file as _read_file
from antsibull_fileutils.io import write_file as _write_file

from .. import app_context
from ..logging import get_module_logger

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

mlog = get_module_logger(__name__)


async def copy_file(
    source_path: StrOrBytesPath, dest_path: StrOrBytesPath, check_content: bool = True
) -> None:
    """
    Copy content from one file to another.

    :arg source_path: Source path. Must be a file.
    :arg dest_path: Destination path.
    :kwarg check_content: If ``True`` (default) and ``lib_ctx.file_check_content > 0`` and the
        destination file exists, first check whether source and destination are potentially equal
        before actually copying,
    """
    lib_ctx = app_context.lib_ctx.get()
    await _copy_file(
        source_path,
        dest_path,
        check_content=check_content,
        file_check_content=lib_ctx.file_check_content,
        chunksize=lib_ctx.chunksize,
    )


async def write_file(filename: StrOrBytesPath, content: str) -> None:
    lib_ctx = app_context.lib_ctx.get()
    await _write_file(filename, content, file_check_content=lib_ctx.file_check_content)


async def read_file(filename: StrOrBytesPath, encoding: str = "utf-8") -> str:
    return await _read_file(filename, encoding=encoding)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to help with hashing."""

from __future__ import annotations

import typing as t
from collections.abc import Mapping

from antsibull_fileutils.hashing import verify_a_hash as _verify_a_hash
from antsibull_fileutils.hashing import verify_hash as _verify_hash

from .. import app_context

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


async def verify_hash(
    filename: StrOrBytesPath,
    hash_digest: str,
    algorithm: str = "sha256",
    algorithm_kwargs: dict[str, t.Any] | None = None,
) -> bool:
    """
    Verify whether a file has a given sha256sum.

    :arg filename: The file to verify the sha256sum of.
    :arg hash_digest: The hash that is expected.
    :kwarg algorithm: The hash algorithm to use.  This must be present in hashlib on this
        system.  The default is 'sha256'
    :kwarg algorithm_kwargs: Parameters to provide to the hash algorithm's constructor.
    :returns: True if the hash matches, otherwise False.
    """
    ctx = app_context.lib_ctx.get()
    return await _verify_hash(
        filename,
        hash_digest,
        algorithm=algorithm,
        algorithm_kwargs=algorithm_kwargs,
        chunksize=ctx.chunksize,
    )


async def verify_a_hash(
    filename: StrOrBytesPath, hash_digests: Mapping[str, str]
) -> bool:
    """
    Verify whether a file has a given hash, given a set of digests with different algorithms.
    Will only test trustworthy hashes and return ``False`` if none matches.

    :arg filename: The file to verify the hash of.
    :arg hash_digest: A mapping of hash types to digests.
    :returns: True if the hash matches, otherwise False.
    """
    ctx = app_context.lib_ctx.get()
    return await _verify_a_hash(filename, hash_digests, chunksize=ctx.chunksize)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Entrypoint to the antsibull-docs script."""

from __future__ import annotations

import typing as t

from antsibull_core.subprocess_util import log_run

try:
    # We fallback to /usr/bin/getfacl so we can ignore failure to import this
    import posix1e  # type: ignore[import] # pyre-ignore[21]
except ImportError:
    posix1e = None

if t.TYPE_CHECKING:
    from _typeshed import StrPath


class UnableToCheck(Exception):
    """No library or command support to check this."""


class CheckFailure(Exception):
    """We checking failed unexpectedly."""


def _get_acls(path: StrPath) -> str:
    """Return the acls for a given file or raise an exception."""
    acls = None
    if posix1e:
        acl = posix1e.ACL(file=path)
        acls = acl.to_any_text(options=posix1e.TEXT_NUMERIC_IDS).decode("utf-8")
    else:
        try:
            acls = log_run(["getfacl", path, "-n"]).stdout
        except FileNotFoundError:
            pass
        except Exception as e:
            # pylint:disable-next=raise-missing-from
            raise CheckFailure(f"Error while trying to get acls for {path}: {e}")

    if not acls:
        raise UnableToCheck(f"No way to determine acls for {path}")

    return acls


def writable_via_acls(path: StrPath, euid: int) -> bool:
    """
    Check whether acls gives someone other than the user write access to a path.

    :arg path: Pathname to check.
    :arg euid: userid to identify the user by.
    :returns: True if acls give someone other than the user path write access to the path.
    """
    # Check acls if available
    acls = _get_acls(path)

    mask_has_write = True
    principal_has_write = False

    # Strip comments and blank lines.  Everything else is an acl
    for acl in (a for a in acls.splitlines() if not a.startswith("#") and a):
        type_, principal, permissions = acl.rsplit(":", 2)
        # default acls have the same issues as acls directly on the directory
        if type_.startswith("default:"):
            dummy_, type_ = type_.split(":", 1)

        # All the safe things:

        # ACL is for the user themselves
        if type_ == "user" and (not principal or principal == euid):
            continue

        if type_ == "mask":
            # The mask does not allow acls to set write
            if "w" not in permissions:
                mask_has_write = False
                break
            # No other mask values affect safety
            continue

        # The acl doesn't grant write permission
        if "w" not in permissions:
            continue

        # Okay, the principal has write in this case
        principal_has_write = True

    if mask_has_write and principal_has_write:
        return True
    return False



# Copyright (C) 2018 Julien Hartmann
# This program is distributed under the MIT license, a copy of which you should
# have receveived along with it. If not, see <https://opensource.org/licenses/MIT>.
#
# SPDX-License-Identifier: MIT
#
# https://github.com/solid-abstractions/python/tree/master/testing-aiohttp-clients

import contextlib
import datetime
import ssl
import tempfile

import pytest


class TemporaryCertificate:
    """Certificate representation used by :func:`ssl_certificate` fixture"""

    def __enter__(self):
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.hazmat.primitives.asymmetric import rsa

        subject = issuer = x509.Name(
            [x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, "localhost")]
        )

        with contextlib.ExitStack() as stack:
            key = rsa.generate_private_key(
                public_exponent=65537, key_size=2048, backend=default_backend()
            )

            key_file = stack.enter_context(tempfile.NamedTemporaryFile())
            key_file.write(
                key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )
            key_file.flush()

            cert = (
                x509.CertificateBuilder()
                .subject_name(subject)
                .issuer_name(issuer)
                .public_key(key.public_key())
                .serial_number(x509.random_serial_number())
                .not_valid_before(datetime.datetime.utcnow())
                .not_valid_after(
                    datetime.datetime.utcnow() + datetime.timedelta(days=1)
                )
                .add_extension(
                    x509.SubjectAlternativeName(
                        [
                            x509.DNSName("localhost"),
                            x509.DNSName("127.0.0.1"),
                        ]
                    ),
                    critical=False,
                )
                .sign(key, hashes.SHA256(), default_backend())
            )

            cert_file = stack.enter_context(tempfile.NamedTemporaryFile())
            cert_file.write(cert.public_bytes(serialization.Encoding.PEM))
            cert_file.flush()

            self._key_file, self._cert_file = key_file, cert_file
            stack.pop_all()
        return self

    def __exit__(self, exc, exc_type, tb):
        self._key_file.close()
        self._cert_file.close()

    def load_verify(self, context):
        """Load the certificate for verification purposes.

        :param ssl.SSLContext context: a SSL context that will be associated with the server.
        """
        context.load_verify_locations(cafile=self._cert_file.name)

    def client_context(self):
        """A client-side SSL context accepting the certificate, and no others"""
        context = ssl.SSLContext()
        context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        self.load_verify(context)
        return context

    def server_context(self):
        """A server-side SSL context using the certificate"""
        context = ssl.SSLContext()
        context.load_cert_chain(self._cert_file.name, keyfile=self._key_file.name)
        return context


@pytest.fixture(scope="session")
def ssl_certificate():
    """Self-signed certificate fixture, used for local server tests."""
    with TemporaryCertificate() as certificate:
        yield certificate



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to deal with collections on the local system"""

from __future__ import annotations

import asyncio
import os
import typing as t
from collections.abc import Sequence

from .subprocess_util import async_log_run

if t.TYPE_CHECKING:
    from _typeshed import StrPath


class CollectionFormatError(Exception):
    pass


async def install_together(
    collection_tarballs: Sequence[StrPath], ansible_collections_dir: StrPath
) -> None:
    installers = []
    for pathname in collection_tarballs:
        basename = os.path.basename(
            pathname  # pyre-ignore[6]: basename() accepts path-like object
        )
        namespace, collection, _dummy = basename.split("-", 2)
        collection_dir = os.path.join(ansible_collections_dir, namespace, collection)
        # Note: mkdir -p equivalent is okay because we created package_dir ourselves as a directory
        # that only we can access
        os.makedirs(collection_dir, mode=0o700, exist_ok=False)
        installers.append(
            asyncio.create_task(
                async_log_run(["tar", "-xf", pathname, "-C", collection_dir])
            )
        )

    await asyncio.gather(*installers)


async def install_separately(
    collection_tarballs: Sequence[StrPath], collection_dir: StrPath
) -> list[str]:
    installers = []
    collection_dirs: list[str] = []

    if not collection_tarballs:
        return collection_dirs

    for pathname in collection_tarballs:
        filename = os.path.basename(
            pathname  # pyre-ignore[6]: basename() accepts path-like object
        )
        namespace, collection, version_ext = filename.split("-", 2)
        version = None
        for ext in (".tar.gz",):
            # Note: If galaxy allows other archive formats, add their extensions here
            ext_start = version_ext.find(ext)
            if ext_start != -1:
                version = version_ext[:ext_start]
                break
        else:
            raise CollectionFormatError(
                "Collection filename was in an unexpected" f" format: {filename}"
            )

        package_dir = os.path.join(
            collection_dir,
            f"ansible-collections-{namespace}.{collection}-{version}",
        )
        os.mkdir(package_dir, mode=0o700)
        collection_dirs.append(package_dir)

        collection_dir = os.path.join(
            package_dir, "ansible_collections", namespace, collection
        )
        # Note: this is okay because we created package_dir ourselves as a directory
        # that only we can access
        os.makedirs(collection_dir, mode=0o700, exist_ok=False)

        installers.append(
            asyncio.create_task(
                async_log_run(["tar", "-xf", pathname, "-C", collection_dir])
            )
        )

    await asyncio.gather(*installers)

    return collection_dirs



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Toshio Kuratomi
"""Pydantic validators."""

from __future__ import annotations

import os.path


def convert_none(value):
    """
    Convert strings to Python None.

    When a setting is set to None in a config file, it could be the string "None" or "Null".
    This validator will convert those strings to  python None.
    """
    if isinstance(value, str) and value.lower() in ("none", "null"):
        value = None
    return value


def _is_truthy_int(value):
    if value == 0:
        return False
    return True


def convert_bool(value):
    """
    Convert strings to Python True/False.

    True and False values may be specified in config files and the command line as strings.  This
    validator will convert a set of predetermined strings to Python True and False.
    """
    if isinstance(value, str):
        if value.lower() in ("false", "no", "n", "f", ""):
            value = False
        elif value.lower() in ("true", "yes", "y", "t"):
            value = True
        else:
            try:
                value = int(value)
            # Any failure to convert to int just means the string does not map to a number.  We can
            # safely ignore that.
            except Exception:  # pylint: disable=broad-except
                pass
            else:
                value = _is_truthy_int(value)

    elif isinstance(value, int):
        value = _is_truthy_int(value)

    return value


def convert_path(value):
    """
    Expand `~` and environment variables in strings. Also convert strings like `None` and `Null`
    to None.
    """
    value = convert_none(value)

    if isinstance(value, str):
        value = os.path.expandvars(os.path.expanduser(value))

    return value



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

import os.path
import sys
from unittest import mock

import pytest

from antsibull_core import subprocess_util
from antsibull_core.venv import FakeVenvRunner, VenvRunner, get_clean_environment


def test_venv_clean_env(monkeypatch):
    monkeypatch.setenv("PYTHONPATH", "/jfjfjfjfjfjfjfjfj")
    assert "PYTHONPATH" not in get_clean_environment()


def test_venv_run_init(tmp_path):
    with mock.patch("antsibull_core.subprocess_util.async_log_run") as log_run:
        runner = VenvRunner("asdfgh", tmp_path)
        assert runner.name == "asdfgh"
        assert runner.top_dir == tmp_path
        assert runner.venv_dir == str(tmp_path / "asdfgh")
        pip = os.path.join(runner.venv_dir, "bin", "pip")
        log_run.assert_called_once_with(
            [pip, "install", "--upgrade", "pip"],
            None,
            None,
            "debug",
            True,
            errors="strict",
            env=get_clean_environment(),
        )


def test_venv_log_run_error(tmp_path):
    runner = VenvRunner("zxcvb", tmp_path)
    echo = os.path.join(runner.venv_dir, "bin", "echo")
    with pytest.raises(ValueError, match=rf"^{echo!r} does not exist!"):
        runner.log_run(["echo", "This won't work!"])


def test_venv_log_run_error2(tmp_path):
    runner = VenvRunner("zxcvb", tmp_path)
    echo = "/usr/bin/echo"
    with pytest.raises(ValueError, match=rf"^{echo!r} must not be an absolute path!"):
        runner.log_run([echo, "This also won't work!"])


def test_fake_venv_python():
    with mock.patch.object(
        subprocess_util, "async_log_run", wraps=subprocess_util.async_log_run
    ) as log_run:
        args = ["python", "-c", 'import antsibull_core; print("Hello, world!")']
        p = FakeVenvRunner().log_run(args)
        assert p.stdout == "Hello, world!\n"

        args = args.copy()
        args[0] = sys.executable
        log_run.assert_called_once_with(
            args,
            None,
            None,
            "debug",
            True,
            errors="strict",
        )



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functionality to work with a venv."""

from __future__ import annotations

import asyncio
import os
import sys
import venv
from collections.abc import Sequence
from typing import TYPE_CHECKING, NoReturn

from antsibull_core import subprocess_util

if TYPE_CHECKING:
    import subprocess
    from logging import Logger as StdLogger

    from _typeshed import StrPath
    from twiggy.logger import Logger as TwiggyLogger  # type: ignore[import]

    from antsibull_core.logging import Logger


def get_clean_environment() -> dict[str, str]:
    env = os.environ.copy()
    try:
        del env["PYTHONPATH"]
    except KeyError:
        # We just wanted to make sure there was no PYTHONPATH set...
        # all python libs will come from the venv
        pass
    return env


class VenvRunner:
    """
    Makes running a command in a venv easy.

    Combines venv functionality with ``antsibull_core.subprocess_util``.

    .. seealso::
        * :python:mod:`venv`
    """

    name: str
    top_dir: StrPath
    venv_dir: str

    def __init__(self, name: str, top_dir: StrPath) -> None:
        """
        Create a venv.

        :arg name: Name of the venv.
        :arg top_dir: Directory the venv will be created inside of.
        """
        self.name = name
        self.top_dir = top_dir
        self.venv_dir: str = os.path.join(top_dir, name)
        venv.create(self.venv_dir, clear=True, symlinks=True, with_pip=True)

        # Upgrade pip to the latest version.
        # Note that cryptography stopped building manylinux1 wheels (the only ship manylinux2010) so
        # we need pip19+ in order to work now.  RHEL8 and Ubuntu 18.04 contain a pip that's older
        # than that so we must upgrade to something even if it's not latest.

        self.log_run(["pip", "install", "--upgrade", "pip"])

    def install_package(self, package_name: str) -> subprocess.CompletedProcess:
        """
        Install a python package into the venv.

        :arg package_name: This can be a bare package name or a path to a file.  It's passed
            directly to :command:`pip install`.
        :returns: An :obj:`subprocess.CompletedProcess` for the pip output.
        """
        return self.log_run(["pip", "install", package_name])

    async def async_log_run(
        self,
        args: Sequence[StrPath],
        logger: TwiggyLogger | StdLogger | Logger | None = None,
        stdout_loglevel: str | subprocess_util.OutputCallbackType | None = None,
        stderr_loglevel: str | subprocess_util.OutputCallbackType | None = "debug",
        check: bool = True,
        *,
        errors: str = "strict",
        **kwargs,
    ) -> subprocess.CompletedProcess[str]:
        """
        This method asynchronously runs a command in a subprocess and logs
        its output. It calls `antsibull_core.subprocess_util.async_log_run` to
        do the heavy lifting. `args[0]` must be a filename that's installed in
        the venv. If it's not, a `ValueError` will be raised.
        """
        kwargs.setdefault("env", get_clean_environment())
        basename = args[0]
        if os.path.isabs(basename):
            raise ValueError(f"{basename!r} must not be an absolute path!")
        path = os.path.join(self.venv_dir, "bin", basename)
        if not os.path.exists(path):
            raise ValueError(f"{path!r} does not exist!")
        args = [path, *args[1:]]
        return await subprocess_util.async_log_run(
            args,
            logger,
            stdout_loglevel,
            stderr_loglevel,
            check,
            errors=errors,
            **kwargs,
        )

    def log_run(
        self,
        args: Sequence[StrPath],
        logger: TwiggyLogger | StdLogger | Logger | None = None,
        stdout_loglevel: str | subprocess_util.OutputCallbackType | None = None,
        stderr_loglevel: str | subprocess_util.OutputCallbackType | None = "debug",
        check: bool = True,
        *,
        errors: str = "strict",
        **kwargs,
    ) -> subprocess.CompletedProcess[str]:
        """
        See :method:`async_log_run`
        """
        return asyncio.run(
            self.async_log_run(
                args,
                logger,
                stdout_loglevel,
                stderr_loglevel,
                check,
                errors=errors,
                **kwargs,
            )
        )


class FakeVenvRunner:
    """
    Simply runs commands.

    .. seealso::
        * :python:mod:`venv`
    """

    async def async_log_run(
        self,
        args: Sequence[StrPath],
        logger: TwiggyLogger | StdLogger | Logger | None = None,
        stdout_loglevel: str | None = None,
        stderr_loglevel: str | None = "debug",
        check: bool = True,
        *,
        errors: str = "strict",
        **kwargs,
    ) -> subprocess.CompletedProcess[str]:
        """
        This method asynchronously runs a command in a subprocess and logs its
        output.
        It works the same as `antsibull_core.subprocess_util.async_log_run`,
        but 'python' will be replaced by `sys.executable`.
        """
        if args and args[0] == "python":
            args = [sys.executable, *args[1:]]
        return await subprocess_util.async_log_run(
            args,
            logger,
            stdout_loglevel,
            stderr_loglevel,
            check,
            errors=errors,
            **kwargs,
        )

    def log_run(
        self,
        args: Sequence[StrPath],
        logger: TwiggyLogger | StdLogger | Logger | None = None,
        stdout_loglevel: str | None = None,
        stderr_loglevel: str | None = "debug",
        check: bool = True,
        *,
        errors: str = "strict",
        **kwargs,
    ) -> subprocess.CompletedProcess[str]:
        """
        See :method:`async_log_run`
        """
        return asyncio.run(
            self.async_log_run(
                args,
                logger,
                stdout_loglevel,
                stderr_loglevel,
                check,
                errors=errors,
                **kwargs,
            )
        )

    @staticmethod
    def install_package(package_name: str) -> NoReturn:
        """
        This raises a NotImplementedError and only exists for parity with
        `VenvRunner`.

        """
        raise NotImplementedError



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Argument parsing helpers."""

from __future__ import annotations

import argparse
import os.path
from importlib import metadata


class InvalidArgumentError(Exception):
    """A problem parsing or validating a command line argument."""


def get_toplevel_parser(
    package: str,
    *,
    package_version: str | None = None,
    program_name: str | None = None,
    **kwargs,
) -> argparse.ArgumentParser:
    """
    Create a toplevel argument parser with options common across all scripts.

    :arg package: The Python package containing this CLI program.
    :arg package_version: If provided, use this instead of the version for ``package``
        provided by ``importlib.metadata.version()``.
    :arg program_name: If provided, show a more concrete description for this program.
    :args kwargs: This function takes any keyword arguments and passes them directly on to
        the :class:`argparse.ArgumentParser` constructor.
    :returns: :class:`argparse.ArgumentParser` with common script arguments added.
    """
    if package_version is None:
        try:
            package_version = metadata.version(package)
        except metadata.PackageNotFoundError:
            # If there's no metadata found, assume we're running from source
            package_version = "source"

    toplevel_parser = argparse.ArgumentParser(**kwargs)
    toplevel_parser.add_argument(
        "--version",
        action="version",
        version=package_version,
        help=(
            f"Print the {program_name} version"
            if program_name
            else "Print the program's version"
        ),
    )
    toplevel_parser.add_argument(
        "--config-file",
        default=[],
        action="append",
        help="Specify one or more config files to use to configure the"
        " program. If more than one are specified, keys from later"
        " config files override keys from earlier ones.",
    )
    return toplevel_parser


def normalize_toplevel_options(args: argparse.Namespace) -> None:
    """
    Normalize and validate the common cli arguments.

    :arg args: The argparse parsed arguments.  The arguments added by the common parser will be
        validated and normalized.

    .. warning:: This function operates by side effect.

        Any normalization needed will be applied directly to ``args``.
    """
    for conf_file in args.config_file:
        if not os.path.isfile(conf_file):
            raise InvalidArgumentError(
                f"The user specified config file, {conf_file}," " must exist."
            )



# This is a copy of BooleanOptionalAction from the Python-3.9 code.
# https://github.com/python/cpython/blob/41b223d29cdfeb1f222c12c3abaccc3bc128f5e7/Lib/argparse.py#L856
# It is licensed under the Python Software Foundation License:
# https://github.com/python/cpython/blob/main/LICENSE
#
# SPDX-FileCopyrightText: 2001-2020 Python Software Foundation
# SPDX-License-Identifier: PSF-2.0
#
# A bug that affects our usage of BoolanOptionalAction was discovered (it displays default: SUPPRESS
# in the help message).  I have added a bugix for that here and will open a PR upstream once
# a maintainer answers a question for me:
# https://github.com/python/cpython/pull/17447/files#r666506839

# The intention is that this is used from antsibull.compat.BooleanOptionalAction rather than from
# this file. That way we use the version in the python stdlib if it is available and use this
# code if it is not.

# pylint:disable=missing-module-docstring,redefined-builtin

from argparse import OPTIONAL, SUPPRESS, ZERO_OR_MORE, Action


def _add_default_to_help_string(action):
    """
    Add the default value to the option help message.

    ArgumentDefaultsHelpFormatter and BooleanOptionalAction both want to add
    the default value to the help message when it isn't already present.  This
    code will do that, detecting cornercases to prevent duplicates or cases
    where it wouldn't make sense to the end user.
    """
    help = action.help

    if "%(default)" not in action.help:
        if action.default is not SUPPRESS:
            defaulting_nargs = [OPTIONAL, ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                help += " (default: %(default)s)"
    return help


class BooleanOptionalAction(Action):
    def __init__(
        self,
        option_strings,
        dest,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):
        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        # if help is not None and default is not None:
        #     help += f" (default: {default})"

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

        self.help = _add_default_to_help_string(self)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self):
        return " | ".join(self.option_strings)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project

"""Functions for working with the ansible-core package."""

from __future__ import annotations

import ast
import os
import re
import tempfile
import typing as t
from urllib.parse import urljoin

import aiofiles
from antsibull_fileutils.hashing import verify_a_hash
from antsibull_fileutils.io import copy_file
from packaging.version import Version as PypiVer

from . import app_context
from .logging import get_module_logger
from .subprocess_util import async_log_run
from .utils.http import retry_get

if t.TYPE_CHECKING:
    import aiohttp.client
    from _typeshed import StrPath


mlog = get_module_logger(__name__)


class UnknownVersion(Exception):
    """Raised when a requested version does not exist."""


class CannotBuild(Exception):
    """Raised when we can't figure out how to build a package."""


class AnsibleCorePyPiClient:
    """Class to retrieve information about ansible-core from Pypi."""

    def __init__(
        self,
        aio_session: "aiohttp.client.ClientSession",
        pypi_server_url: t.Optional[str] = None,
    ) -> None:
        """
        Initialize the AnsibleCorePypiClient class.

        :arg aio_session: :obj:`aiohttp.client.ClientSession` to make requests to pypi from.
        :kwarg pypi_server_url: URL to the pypi server to use. Defaults to
            ``lib_ctx.get().pypi_url``.
        """
        self.aio_session = aio_session
        if pypi_server_url is None:
            pypi_server_url = str(app_context.lib_ctx.get().pypi_url)
        self.pypi_server_url: str = pypi_server_url

    async def _get_json(self, query_url: str) -> dict[str, t.Any]:
        """
        JSON data from a url with retries and return the data as python data structures.
        """
        async with retry_get(self.aio_session, query_url) as response:
            pkg_info = await response.json()
        return pkg_info

    async def get_release_info(
        self, package_name: t.Optional[str] = None
    ) -> dict[str, t.Any]:
        """
        Retrieve information about releases of the ansible-core package from pypi.

        :arg package_name: Either 'ansible-core' or None.
        :returns: The dict which represents the release info keyed by version number.
            To examine the data structure, use::

                curl https://pypi.org/pypi/ansible-core/json| python3 -m json.tool
        """
        # Retrieve the ansible-core package info from pypi
        package_name = package_name or "ansible-core"
        query_url = urljoin(self.pypi_server_url, f"pypi/{package_name}/json")
        resp = await self._get_json(query_url)
        return resp["releases"]

    async def get_versions(self) -> list[PypiVer]:
        """
        Get the versions of the ansible-core package on pypi.

        :returns: A list of :pypkg:obj:`packaging.versioning.Version`s
            for all the versions on pypi, including prereleases.
        """
        flog = mlog.fields(func="AnsibleCorePyPiClient.get_versions")
        flog.debug("Enter")

        release_info = await self.get_release_info()
        versions = [PypiVer(r) for r in release_info]
        versions.sort(reverse=True)
        flog.fields(versions=versions).info("sorted list of ansible-core versions")

        flog.debug("Leave")
        return versions

    async def get_latest_version(self) -> PypiVer:
        """
        Get the latest version of ansible-core uploaded to pypi.

        :return: A :pypkg:obj:`packaging.versioning.Version` object representing the latest version
            of the package on pypi.  This may be a pre-release.
        """
        versions = await self.get_versions()
        return versions[0]

    # TODO(anyone): Make function less complex
    async def retrieve(  # noqa C901
        self, ansible_core_version: str, download_dir: StrPath
    ) -> str:
        """
        Get the release from pypi.

        :arg ansible_core_version: Version of ansible-core to download.
        :arg download_dir: Directory to download the tarball to.
        :returns: The name of the downloaded tarball.
        """
        package_name = "ansible-core"

        # https://github.com/pypa/setuptools/pull/4286
        # Newer setuptools versions perform dist filename normalization
        tar_filenames = (
            f"{package_name.replace('-', '_')}-{ansible_core_version}.tar.gz",
            f"{package_name}-{ansible_core_version}.tar.gz",
        )
        tar_filename = tar_filenames[0]
        tar_path = os.path.join(download_dir, tar_filename)

        lib_ctx = app_context.lib_ctx.get()
        if lib_ctx.ansible_core_cache and lib_ctx.trust_ansible_core_cache:
            for tar_filename in tar_filenames:
                cached_path = os.path.join(
                    t.cast(str, lib_ctx.ansible_core_cache), tar_filename
                )
                if os.path.isfile(cached_path):
                    tar_path = os.path.join(download_dir, tar_filename)
                    await copy_file(
                        cached_path,
                        tar_path,
                        check_content=False,
                        file_check_content=lib_ctx.file_check_content,
                        chunksize=lib_ctx.chunksize,
                    )
                    return tar_path

        release_info = await self.get_release_info(package_name)

        pypi_url = ""
        digests = {}
        for release in release_info[ansible_core_version]:
            if release["filename"] in tar_filenames:
                tar_filename = release["filename"]
                tar_path = os.path.join(download_dir, tar_filename)
                pypi_url = release["url"]
                digests = release["digests"]
                break
        else:  # for-else: http://bit.ly/1ElPkyg
            raise UnknownVersion(
                f"{package_name} {ansible_core_version} does not"
                f" exist on {self.pypi_server_url}"
            )

        if lib_ctx.ansible_core_cache and "sha256" in digests:
            cached_path = os.path.join(lib_ctx.ansible_core_cache, tar_filename)
            if os.path.isfile(cached_path):
                if await verify_a_hash(
                    cached_path, digests, chunksize=lib_ctx.chunksize
                ):
                    await copy_file(
                        cached_path,
                        tar_path,
                        check_content=False,
                        file_check_content=lib_ctx.file_check_content,
                        chunksize=lib_ctx.chunksize,
                    )
                    return tar_path

        async with retry_get(self.aio_session, pypi_url) as response:
            async with aiofiles.open(tar_path, "wb") as f:
                while chunk := await response.content.read(lib_ctx.chunksize):
                    await f.write(chunk)

        if lib_ctx.ansible_core_cache:
            cached_path = os.path.join(lib_ctx.ansible_core_cache, tar_filename)
            await copy_file(
                tar_path,
                cached_path,
                check_content=False,
                file_check_content=lib_ctx.file_check_content,
                chunksize=lib_ctx.chunksize,
            )

        return tar_path


def _get_source_version(ansible_core_source: StrPath) -> PypiVer:
    with open(
        os.path.join(ansible_core_source, "lib", "ansible", "release.py"),
        "r",
        encoding="utf-8",
    ) as f:
        root = ast.parse(f.read())

    # Find the version of the source
    source_version = None
    # Iterate backwards in case __version__ is assigned to multiple times
    for node in reversed(root.body):
        if isinstance(node, ast.Assign):
            for name in node.targets:
                # These attributes are dynamic so pyre cannot check them
                if name.id == "__version__":  # type: ignore[attr-defined] # pyre-ignore[16]
                    source_version = node.value.s  # type: ignore[attr-defined] # pyre-ignore[16]
                    break

        if source_version:
            break

    if not source_version:
        raise ValueError("Version was not found")

    return PypiVer(source_version)


def _version_is_devel(version: PypiVer) -> bool:
    dev_version = re.compile(".*[.]dev[0-9]+$")
    return bool(dev_version.match(version.public))


def source_is_devel(ansible_core_source: StrPath | None) -> bool:
    """
    :arg ansible_core_source: A path to an Ansible-core checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_core_version``.
    :returns: True if the source looks like it is for the devel branch.
    """
    if ansible_core_source is None:
        return False

    try:
        source_version = _get_source_version(ansible_core_source)
    except Exception:  # pylint:disable=broad-except
        return False

    return _version_is_devel(source_version)


def source_is_correct_version(
    ansible_core_source: StrPath | None, ansible_core_version: PypiVer
) -> bool:
    """
    :arg ansible_core_source: A path to an Ansible-core checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_core_version``.
    :arg ansible_core_version: Version of ansible-core to retrieve.
    :returns: True if the source is for a compatible version at or newer than the requested version
    """
    if ansible_core_source is None:
        return False

    try:
        source_version = _get_source_version(ansible_core_source)
    except Exception:  # pylint:disable=broad-except
        return False

    # If the source is a compatible version of ansible-core and it is the same or more recent than
    # the requested version then allow this.
    if (
        source_version.major == ansible_core_version.major
        and source_version.minor == ansible_core_version.minor
        and source_version.micro >= ansible_core_version.micro
    ):
        return True

    return False


async def checkout_from_git(
    download_dir: StrPath, repo_url: t.Optional[str] = None
) -> str:
    """
    Checkout the ansible-core git repo.

    :arg download_dir: Directory to checkout into.
    :kwarg: repo_url: The URL to the git repo. Defaults to
        ``lib_ctx.get().ansible_core_repo_url``.
    :return: The directory that ansible-core has been checked out to.
    """
    if repo_url is None:
        repo_url = str(app_context.lib_ctx.get().ansible_core_repo_url)
    ansible_core_dir = os.path.join(download_dir, "ansible-core")
    await async_log_run(["git", "clone", repo_url, ansible_core_dir])

    return ansible_core_dir


async def create_sdist(source_dir: StrPath, dest_dir: StrPath) -> str:
    """
    Create an sdist for the python package at a given path.

    :arg source_dir: the directory that the python package source is in.
    :arg dest_dir: the directory that the sdist will be written to/
    :returns: path to the sdist.
    """

    # Make a subdir of dest_dir for returning the dist in
    dist_dir_prefix = os.path.join(
        os.path.basename(
            source_dir  # pyre-ignore[6]: basename() accepts path-like object
        )
    )
    dist_dir = tempfile.mkdtemp(prefix=dist_dir_prefix, dir=dest_dir)

    try:
        await async_log_run(
            ["python", "-m", "build", "--sdist", "--outdir", dist_dir, source_dir],
            stderr_loglevel="warning",
        )
    except Exception as e:
        raise CannotBuild(  # pylint:disable=raise-missing-from
            f"Building {source_dir} failed: {e}"
        )

    dist_files = [f for f in os.listdir(dist_dir) if f.endswith("tar.gz")]
    if len(dist_files) != 1:
        if not dist_files:
            raise CannotBuild(f"Building {source_dir} did not create a tar.gz")

        raise CannotBuild(
            f"Building {source_dir} created more than one tar.gz files which is not"
            " yet supported."
        )

    return os.path.join(dist_dir, dist_files[0])


async def get_ansible_core(
    aio_session: "aiohttp.client.ClientSession",
    ansible_core_version: str,
    tmpdir: StrPath,
    ansible_core_source: StrPath | None = None,
) -> str:
    """
    Create an ansible-core directory of the requested version.

    :arg aio_session: :obj:`aiohttp.client.ClientSession` to make http requests with.
    :arg ansible_core_version: Version of ansible-core to retrieve.  If it is the special string
        ``@devel``, then we will retrieve ansible-core from its git repository.  If it is the
        special string ``@latest``, then we will retrieve the latest version from pypi.
    :arg tmpdir: Temporary directory use as a scratch area for downloading to and the place that the
        ansible-core directory should be placed in.
    :kwarg ansible_core_source: If given, a path to an ansible-core checkout or expanded sdist.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_core_version``.
    """
    if ansible_core_version == "@devel":
        # is the source usable?
        if source_is_devel(ansible_core_source):
            # source_is_devel() protects against this.  This assert is to inform the type checker
            assert ansible_core_source is not None
            source_location: StrPath = ansible_core_source

        else:
            source_location = await checkout_from_git(tmpdir)

        # Create an sdist from the source that can be installed
        install_file = await create_sdist(source_location, tmpdir)
    else:
        pypi_client = AnsibleCorePyPiClient(aio_session)
        ansible_core_pypi_version: PypiVer
        if ansible_core_version == "@latest":
            ansible_core_pypi_version = await pypi_client.get_latest_version()
        else:
            ansible_core_pypi_version = PypiVer(ansible_core_version)

        # is the source the asked for version?
        if source_is_correct_version(ansible_core_source, ansible_core_pypi_version):
            assert ansible_core_source is not None
            # Create an sdist from the source that can be installed
            install_file = await create_sdist(ansible_core_source, tmpdir)
        else:
            install_file = await pypi_client.retrieve(
                ansible_core_pypi_version.public, tmpdir
            )

    return install_file



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import hashlib

import aiohttp
import pytest

from antsibull_core.galaxy import (
    CollectionDownloader,
    GalaxyClient,
    GalaxyContext,
    GalaxyVersion,
)

KNOWN_CG_VERSIONS = [
    "6.4.0",
    "6.3.0",
    "6.2.0",
    "6.1.0",
    "6.0.1",
    "6.0.0",
    "6.0.0-a1",
    "5.8.6",
    "5.8.5",
    "5.8.4",
    "5.8.3",
    "5.8.2",
    "5.8.1",
    "5.8.0",
    "5.7.0",
    "5.6.0",
    "5.5.0",
    "5.4.0",
    "5.3.0",
    "5.2.0",
    "5.1.1",
    "5.1.0",
    "5.0.2",
    "5.0.1",
    "5.0.0",
    "5.0.0-a1",
    "4.8.9",
    "4.8.8",
    "4.8.7",
    "4.8.6",
    "4.8.5",
    "4.8.4",
    "4.8.3",
    "4.8.2",
    "4.8.1",
    "4.8.0",
    "4.7.0",
    "4.6.1",
    "4.6.0",
    "4.5.0",
    "4.4.0",
    "4.3.0",
    "4.2.0",
    "4.1.0",
    "4.0.2",
    "4.0.1",
    "4.0.0",
    "3.8.10",
    "3.8.9",
    "3.8.8",
    "3.8.7",
    "3.8.6",
    "3.8.5",
    "3.8.4",
    "3.8.3",
    "3.8.2",
    "3.8.1",
    "3.8.0",
    "3.7.0",
    "3.6.0",
    "3.5.0",
    "3.4.0",
    "3.3.2",
    "3.3.1",
    "3.3.0",
    "3.2.0",
    "3.1.0",
    "3.0.2",
    "3.0.1",
    "3.0.0",
    "2.5.9",
    "2.5.8",
    "2.5.7",
    "2.5.6",
    "2.5.5",
    "2.5.4",
    "2.5.3",
    "2.5.2",
    "2.5.1",
    "2.5.0",
    "2.4.0",
    "2.3.0",
    "2.2.0",
    "2.1.1",
    "2.1.0",
    "2.0.1",
    "2.0.0",
    "1.3.14",
    "1.3.13",
    "1.3.12",
    "1.3.11",
    "1.3.10",
    "1.3.9",
    "1.3.8",
    "1.3.7",
    "1.3.6",
    "1.3.5",
    "1.3.4",
    "1.3.3",
    "1.3.2",
    "1.3.1",
    "1.3.0",
    "1.2.0",
    "1.1.0",
    "1.0.0",
    "0.3.0-experimental.meta.redirects-3",
    "0.3.0-experimental.meta.redirects-2",
    "0.3.0-experimental.meta.redirects",
    "0.2.1",
    "0.2.0",
    "0.1.4",
    "0.1.1",
]


async def galaxy_client_test(
    aio_session: aiohttp.ClientSession,
    context: GalaxyContext,
    tmp_path_factory,
    skip_versions_test: bool = False,
    skip_download_test: bool = False,
) -> None:
    client = GalaxyClient(aio_session, context=context)

    # Check collection info
    cg_info = await client.get_info("community.general")
    print(cg_info)
    assert cg_info["name"] == "general"
    assert cg_info["deprecated"] is False

    # Check info on specific collection version
    cg_version = await client.get_release_info("community.general", "6.0.0")
    cg_version.pop("files", None)  # this is huge for Galaxy v3 API
    cg_version.pop("metadata", None)  # this is large for Galaxy v3 API
    print(cg_version)
    assert "download_url" in cg_version
    assert cg_version["artifact"]["filename"] == "community-general-6.0.0.tar.gz"
    assert cg_version["artifact"]["size"] == 2285782
    expected = "7c7ec856355078577b520f7432645754d75ad8e74a46e84d1ffee8fad80efc5c"
    assert cg_version["artifact"]["sha256"] == expected
    assert cg_version["namespace"]["name"] == "community"
    assert cg_version["collection"]["name"] == "general"
    assert cg_version["version"] == "6.0.0"

    # Check collection list
    if not skip_versions_test:
        cg_versions = await client.get_versions("community.general")
        print(cg_versions)
        for known_version in KNOWN_CG_VERSIONS:
            assert known_version in cg_versions

    # Download collection
    if not skip_download_test:
        download_path = tmp_path_factory.mktemp("download")
        cache_path = tmp_path_factory.mktemp("cache")
        downloader = CollectionDownloader(
            aio_session,
            str(download_path),
            collection_cache=str(cache_path),
            context=context,
        )
        path = await downloader.download("community.dns", "0.1.0")
        with open(path, "rb") as f:
            data = f.read()
        length = len(data)
        print(length)
        assert length == 133242
        m = hashlib.sha256()
        m.update(data)
        digest = m.hexdigest()
        expected = "2de9d40940536e65b35995a3f58dea7776de3f23f1a7ab865e0b3b8482d746b5"
        assert digest == expected

        # Try again, should hit cache
        path2 = await downloader.download("community.dns", "0.1.0")
        with open(path2, "rb") as f:
            data2 = f.read()
        assert path == path2
        assert data == data2


@pytest.mark.asyncio
async def test_galaxy_v2(tmp_path_factory):
    galaxy_url = "https://old-galaxy.ansible.com"
    async with aiohttp.ClientSession() as aio_session:
        context = await GalaxyContext.create(aio_session, galaxy_url)
        assert context.version == GalaxyVersion.V2
        assert context.base_url == galaxy_url + "/api/v2/"
        await galaxy_client_test(aio_session, context, tmp_path_factory)


@pytest.mark.asyncio
async def test_galaxy_v3(tmp_path_factory):
    galaxy_url = "https://galaxy.ansible.com"
    async with aiohttp.ClientSession() as aio_session:
        context = await GalaxyContext.create(aio_session, galaxy_url)
        assert context.version == GalaxyVersion.V3
        assert context.base_url == galaxy_url + "/api/v3/"
        await galaxy_client_test(aio_session, context, tmp_path_factory)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import pytest

from antsibull_core import dependency_files as df

PIECES = """
community.general
# This is a comment
community.aws
    # Supported by ansible
    ansible.posix
    ansible.windows
# Third parties
purestorage.flasharray
"""

PARSED_PIECES = [
    "community.general",
    "community.aws",
    "ansible.posix",
    "ansible.windows",
    "purestorage.flasharray",
]

DEPS = """
_ansible_version: 2.10.5
# this is a comment
_ansible_core_version: 2.10.4
    # supported by ansible
    ansible.netcommon: 1.4.1
    ansible.posix: 1.1.1
    ansible.windows: 1.3.0
# third parties
dellemc.os10: 1.0.2
"""

PARSED_DEPS_ANSIBLE_VERSION = "2.10.5"
PARSED_DEPS_ANSIBLE_CORE_VERSION = "2.10.4"
PARSED_DEPS_DEPS = {
    "ansible.netcommon": "1.4.1",
    "ansible.posix": "1.1.1",
    "ansible.windows": "1.3.0",
    "dellemc.os10": "1.0.2",
}

DEPS_2 = """
_ansible_version: 2.10.5
# this is a comment
_ansible_core_version: 2.10.4
    # supported by ansible
    ansible.netcommon: 1.4.1
    ansible.posix: 1.1.1
    ansible.windows: 1.3.0
# third parties
dellemc.os10: 1.0.2
"""

PARSED_DEPS_2_ANSIBLE_VERSION = "2.10.5"
PARSED_DEPS_2_ANSIBLE_CORE_VERSION = "2.10.4"
PARSED_DEPS_2_DEPS = {
    "ansible.netcommon": "1.4.1",
    "ansible.posix": "1.1.1",
    "ansible.windows": "1.3.0",
    "dellemc.os10": "1.0.2",
}

BUILD = """
_ansible_version: 2.10
# this is a comment
_ansible_core_version: 2.10.1
    # supported by ansible
    ansible.netcommon: >=1.2.0,<2.0.0
    ansible.posix: >=1.1.0,<2.0.0
    ansible.windows: >=1.0.0,<2.0.0
# third parties
dellemc.os10: >=1.0.0,<1.1.0
"""

PARSED_BUILD_ANSIBLE_VERSION = "2.10"
PARSED_BUILD_ANSIBLE_CORE_VERSION = "2.10.1"
PARSED_BUILD_DEPS = {
    "ansible.netcommon": ">=1.2.0,<2.0.0",
    "ansible.posix": ">=1.1.0,<2.0.0",
    "ansible.windows": ">=1.0.0,<2.0.0",
    "dellemc.os10": ">=1.0.0,<1.1.0",
}

BUILD_2 = """
_ansible_version: 2.10
# this is a comment
_ansible_core_version: 2.10.1
    # supported by ansible
    ansible.netcommon: >=1.2.0,<2.0.0
    ansible.posix: >=1.1.0,<2.0.0
    ansible.windows: >=1.0.0,<2.0.0
# third parties
dellemc.os10: >=1.0.0,<1.1.0
"""

PARSED_BUILD_2_ANSIBLE_VERSION = "2.10"
PARSED_BUILD_2_ANSIBLE_CORE_VERSION = "2.10.1"
PARSED_BUILD_2_DEPS = {
    "ansible.netcommon": ">=1.2.0,<2.0.0",
    "ansible.posix": ">=1.1.0,<2.0.0",
    "ansible.windows": ">=1.0.0,<2.0.0",
    "dellemc.os10": ">=1.0.0,<1.1.0",
}


def test_parse_pieces(tmp_path):
    pieces_filename = tmp_path / "pieces.in"
    with open(pieces_filename, "w") as f:
        f.write(PIECES)
    assert df.parse_pieces_file(pieces_filename) == PARSED_PIECES


def test_deps_file_parse(tmp_path):
    deps_filename = tmp_path / "deps.in"
    with open(deps_filename, "w") as f:
        f.write(DEPS)
    parsed_deps = df.DepsFile(deps_filename).parse()
    assert parsed_deps.ansible_version == PARSED_DEPS_ANSIBLE_VERSION
    assert parsed_deps.ansible_core_version == PARSED_DEPS_ANSIBLE_CORE_VERSION
    assert parsed_deps.deps == PARSED_DEPS_DEPS


def test_deps_file_2_parse(tmp_path):
    deps_filename = tmp_path / "deps.in"
    with open(deps_filename, "w") as f:
        f.write(DEPS_2)
    parsed_deps = df.DepsFile(deps_filename).parse()
    assert parsed_deps.ansible_version == PARSED_DEPS_2_ANSIBLE_VERSION
    assert parsed_deps.ansible_core_version == PARSED_DEPS_2_ANSIBLE_CORE_VERSION
    assert parsed_deps.deps == PARSED_DEPS_2_DEPS


def test_build_file_parse(tmp_path):
    build_filename = tmp_path / "build.in"
    with open(build_filename, "w") as f:
        f.write(BUILD)
    parsed_build = df.DepsFile(build_filename).parse()
    assert parsed_build.ansible_version == PARSED_BUILD_ANSIBLE_VERSION
    assert parsed_build.ansible_core_version == PARSED_BUILD_ANSIBLE_CORE_VERSION
    assert parsed_build.deps == PARSED_BUILD_DEPS


def test_build_file_2_parse(tmp_path):
    build_filename = tmp_path / "build.in"
    with open(build_filename, "w") as f:
        f.write(BUILD_2)
    parsed_build = df.DepsFile(build_filename).parse()
    assert parsed_build.ansible_version == PARSED_BUILD_2_ANSIBLE_VERSION
    assert parsed_build.ansible_core_version == PARSED_BUILD_2_ANSIBLE_CORE_VERSION
    assert parsed_build.deps == PARSED_BUILD_2_DEPS



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""General functions for working with aiohttp."""

from __future__ import annotations

import asyncio
import math
import random
import typing as t
import warnings
from collections.abc import Iterable, Mapping

import aiohttp

from .. import app_context
from ..logging import get_module_logger

mlog = get_module_logger(__name__)


# Since Python 3.11 asyncio.TimeoutError is a deprecated alias of TimeoutError
_AsyncIoTimeoutError = getattr(asyncio, "TimeoutError", TimeoutError)  # pyre-ignore[16]


def _format_call(
    command: str, args: tuple[t.Any, ...], kwargs: Mapping[str, t.Any]
) -> str:
    arguments = [repr(a) for a in args] + [f"{k}={repr(v)}" for k, v in kwargs.items()]
    return f'aio_session.{command}({", ".join(arguments)})'


class RetryGetManager:
    response: aiohttp.ClientResponse | None

    def __init__(
        self,
        aio_session: aiohttp.client.ClientSession,
        args: tuple[t.Any, ...],
        kwargs: Mapping[str, t.Any],
        max_retries: int,
        acceptable_error_codes: Iterable[int],
    ):
        self.aio_session = aio_session
        self.args = args
        self.kwargs = kwargs
        self.max_retries = max_retries
        self.acceptable_error_codes = acceptable_error_codes
        self.call_string = _format_call("get", args, kwargs)
        self.response = None

    async def __aenter__(self) -> aiohttp.ClientResponse:
        flog = mlog.fields(func="RetryGetManager.__aenter__")
        flog.debug("Enter")

        errors = []
        for retry in range(self.max_retries):
            flog.debug("Execute {0}", self.call_string)
            wait_factor: float = 5
            try:
                response = await self.aio_session.get(  # pyre-ignore[16]
                    *self.args, **self.kwargs, timeout=aiohttp.ClientTimeout(total=20)
                )
                status_code = response.status
                flog.debug(f"Status code {status_code}")
                if status_code < 400 or status_code in self.acceptable_error_codes:
                    self.response = response
                    flog.debug("Leave")
                    return response
                status = str(status_code)
                response.close()
            except _AsyncIoTimeoutError:
                status = "timeout"
                wait_factor = 0.5
            except Exception as error:  # pylint:disable=broad-except
                status = str(error)

            errors.append(status)
            failed = retry + 1 == self.max_retries
            warnings.warn(
                f"{self.call_string} failed with status code {status}"
                f'{", finally failed." if failed else ", retrying..."}'
            )
            if failed:
                break

            await asyncio.sleep(
                math.pow(1.5, retry) * wait_factor + (0.5 + random.uniform(0, 1))
            )

        flog.debug("Raise error")
        raise RuntimeError(
            f"Repeated error when calling {self.call_string}: received status codes "
            f'{", ".join(errors)}'
        )

    async def __aexit__(self, exc_type, exc, tb) -> None:
        flog = mlog.fields(func="RetryGetManager.__aexit__")
        flog.debug("Enter")
        response = self.response
        self.response = None
        if response is not None:
            response.close()
        flog.debug("Leave")


def retry_get(
    aio_session: aiohttp.client.ClientSession,
    *args,
    acceptable_error_codes: Iterable[int] | None = None,
    max_retries: int | None = None,
    **kwargs,
) -> t.AsyncContextManager[aiohttp.ClientResponse]:
    # Handle default value for max_retries
    lib_ctx = app_context.lib_ctx.get()
    if max_retries is None:
        max_retries = lib_ctx.max_retries

    # Make sure max_retries is at least 1
    max_retries = max(max_retries, 1)

    return RetryGetManager(
        aio_session, args, kwargs, max_retries, acceptable_error_codes or ()
    )



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import pytest
from packaging.version import Version

import antsibull_core.ansible_core as ac


@pytest.mark.parametrize(
    "version, is_devel",
    [
        ("2.14.0dev0", True),
        ("2.14.0", False),
    ],
)
def test_get_core_package_name_returns_ansible_core(version, is_devel):
    assert ac._version_is_devel(Version(version)) == is_devel



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

import logging as stdlog
from unittest.mock import MagicMock, call

import pytest
import twiggy

import antsibull_core.subprocess_util
from antsibull_core import logging


def test_log_run() -> None:
    logger = MagicMock()
    args = ("bash", "-ec", "echo 123 && echo 456 >&2")
    proc = antsibull_core.subprocess_util.log_run(args, logger)
    assert proc.args == args
    assert proc.returncode == 0
    assert proc.stdout == "123\n"
    assert proc.stderr == "456\n"
    calls = [call(f"Running subprocess: {args}"), call("stderr: 456")]
    assert logger.debug.call_args_list == calls


def test_log_run_multi() -> None:
    logger = MagicMock()
    command = """
    for i in {1..15}; do
        echo "$i: Hello, stdout"
        echo "$i: Hello, stderr" >&2
    done
    """
    args = ("bash", "-ec", command)
    proc = antsibull_core.subprocess_util.log_run(args, logger, "info", "warn")
    assert proc.args == args
    assert proc.returncode == 0

    logger.debug.assert_called_once_with(f"Running subprocess: {args}")
    assert logger.warn.call_count == 15
    assert logger.info.call_count == 15
    expected_out: list[str] = []
    expected_err: list[str] = []
    for index, inp in enumerate(logger.info.call_args_list):
        msg = f"{index+1}: Hello, stdout"
        expected_out.append(msg)
        assert inp == call("stdout: " + msg)
    for index, inp in enumerate(logger.warn.call_args_list):
        msg = f"{index+1}: Hello, stderr"
        expected_err.append(msg)
        assert inp == call("stderr: " + msg)
    assert proc.stdout == "\n".join(expected_out) + "\n"
    assert proc.stderr == "\n".join(expected_err) + "\n"


@pytest.mark.parametrize(
    "count",
    [
        8 * 1024 * 1024 - 1,  # should not trigger long line code
        8 * 1024 * 1024,  # should not trigger long line code
        8 * 1024 * 1024 + 1,
        8 * 1024 * 1024 + 10,
        9 * 1024 * 1024,
    ],
)
def test_log_run_long_line(count: int) -> None:
    args = (
        "sh",
        "-c",
        f"dd if=/dev/zero of=/dev/stdout bs={count} count=1 ; echo ; echo foo",
    )
    proc = antsibull_core.subprocess_util.log_run(args)
    assert proc.args == args
    assert proc.returncode == 0
    assert proc.stdout == ("\u0000" * count) + "\nfoo\n"


def test_log_run_callback() -> None:
    stdout_lines: list[str] = []
    stderr_lines: list[str] = []

    async def add_to_stderr(string: str, /) -> None:
        stderr_lines.append(string)

    antsibull_core.subprocess_util.log_run(
        ["sh", "-c", "echo Never; echo gonna >&2; echo give"],
        None,
        stdout_lines.append,
        add_to_stderr,
    )
    assert stdout_lines == ["Never", "give"]
    assert stderr_lines == ["gonna"]


def test_log_run_brackets_twiggy(capsys: pytest.CaptureFixture) -> None:
    try:
        logging.initialize_app_logging()
        msg = "{abc} {x} }{}"
        args = ("echo", msg)
        antsibull_core.subprocess_util.log_run(
            args, logger=logging.log, stdout_loglevel="error"
        )
        _, stderr = capsys.readouterr()
        assert stderr == f"ERROR:antsibull|stdout: {msg}\n"
    finally:
        logging.log.min_level = twiggy.levels.DISABLED


def test_log_run_percent_logging(capsys: pytest.CaptureFixture) -> None:
    logger = stdlog.Logger("test_logger")
    logger.setLevel(stdlog.WARNING)
    ch = stdlog.StreamHandler()
    ch.setLevel(stdlog.WARNING)
    formatter = stdlog.Formatter("%(levelname)s:%(name)s|%(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    msg = "%s %(abc)s % %%"
    args = ("echo", msg)
    antsibull_core.subprocess_util.log_run(args, logger=logger, stdout_loglevel="error")
    _, stderr = capsys.readouterr()
    assert stderr == f"ERROR:test_logger|stdout: {msg}\n"



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Utilities for dealing with subprocesses
"""

from __future__ import annotations

import asyncio
import subprocess
import sys
from asyncio.exceptions import IncompleteReadError, LimitOverrunError
from collections.abc import Awaitable, Callable, Sequence
from functools import partial
from inspect import isawaitable
from logging import Logger as StdLogger
from typing import TYPE_CHECKING, Any, TypeVar, cast

from twiggy.logger import Logger as TwiggyLogger  # type: ignore[import]

from antsibull_core.logging import Logger, get_module_logger

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath
    from typing_extensions import ParamSpec, TypeAlias

    _T = TypeVar("_T")
    _P = ParamSpec("_P")

mlog = get_module_logger(__name__)

CalledProcessError = subprocess.CalledProcessError

OutputCallbackType: TypeAlias = "Callable[[str], Any] | Callable[[str], Awaitable[Any]]"

stdout_callback = print
stderr_callback = partial(print, file=sys.stderr)


async def _sync_or_async(
    func: Callable[_P, Awaitable[_T]] | Callable[_P, _T],
    /,
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> _T:
    out = func(*args, **kwargs)
    if isawaitable(out):
        return await out
    return cast("_T", out)


async def _stream_log(
    name: str,
    callback: OutputCallbackType | None,
    stream: asyncio.StreamReader,
    errors: str,
) -> str:
    # We do not simply use stream.readline() since it has a line length limit.
    # While we set this limit already to 8 MB (the default is 64 KB), we still
    # want to cover longer lines as well, so we use stream.readuntil('\n')
    # and manually handle the case of longer lines.
    lines = []
    line_parts = []
    sep = b"\n"
    while True:
        try:
            line_parts.append(await stream.readuntil(sep))
        except IncompleteReadError as e:
            line_parts.append(e.partial)
        except LimitOverrunError as e:
            part = await stream.read(e.consumed)
            line_parts.append(part)
            if part:
                continue

        line = b"".join(line_parts)
        line_parts.clear()
        if not line:
            break
        text = line.decode("utf-8", errors=errors)
        if callback:
            await _sync_or_async(callback, f"{name}{text.strip()}")
        lines.append(text)
    return "".join(lines)


def _get_log_func_and_prefix(
    name: str, loglevel: str | OutputCallbackType | None, logger: Any
) -> tuple[OutputCallbackType | None, str]:
    logfunc: Callable[[str], Any] | None = None
    log_prefix = f"{name}: "
    if loglevel:
        if callable(loglevel):
            logfunc = loglevel
            log_prefix = ""
        else:
            # fmt: off
            func = getattr(logger, loglevel)
            if isinstance(logger, (TwiggyLogger, Logger)):
                def logfunc(string: str, /):
                    func("{0}", string)
            elif isinstance(logger, StdLogger):
                def logfunc(string: str, /):
                    func("%s", string)
            else:
                logfunc = func
            # fmt: on

    return logfunc, log_prefix


async def async_log_run(
    args: Sequence[StrOrBytesPath],
    logger: TwiggyLogger | StdLogger | Logger | None = None,
    stdout_loglevel: str | OutputCallbackType | None = None,
    stderr_loglevel: str | OutputCallbackType | None = "debug",
    check: bool = True,
    *,
    errors: str = "strict",
    **kwargs,
) -> subprocess.CompletedProcess[str]:
    """
    Asynchronously run a command in a subprocess and log its output.
    The command's stdout and stderr are always captured.
    For some usecases, you may still need to call
    asyncio.create_subprocess_exec() directly to have more control.

    :param args: Command to run
    :param logger:
        Logger in which to log the command. Can be a `twiggy.logger.Logger` or
        a stdlib `logger.Logger`.
    :param stdout_loglevel:
        Which level to use to log stdout or a generic callback function.
        `None` disables logging.
    :param stderr_loglevel:
        Which level to use to log stdout or a generic callback function.
        `None` disables logging.
    :param check:
        Whether to raise a `subprocess.CalledProcessError` when the
        command returns a non-zero exit code
    :param errors:
        How to handle UTF-8 decoding errors. Default is ``strict``.
    """
    logger = logger or mlog
    stdout_logfunc, stdout_log_prefix = _get_log_func_and_prefix(
        "stdout", stdout_loglevel, logger
    )
    stderr_logfunc, stderr_log_prefix = _get_log_func_and_prefix(
        "stderr", stderr_loglevel, logger
    )
    logger.debug(f"Running subprocess: {args!r}")
    kwargs["stdout"] = asyncio.subprocess.PIPE
    kwargs["stderr"] = asyncio.subprocess.PIPE
    kwargs["limit"] = 2**23  # Increase line length limit to 8 MB (the default is 64k)
    proc = await asyncio.create_subprocess_exec(*args, **kwargs)
    stdout, stderr = await asyncio.gather(
        # proc.stdout and proc.stderr won't be None with PIPE, hence the cast()
        asyncio.create_task(
            _stream_log(
                stdout_log_prefix,
                stdout_logfunc,
                cast(asyncio.StreamReader, proc.stdout),
                errors,
            )
        ),
        asyncio.create_task(
            _stream_log(
                stderr_log_prefix,
                stderr_logfunc,
                cast(asyncio.StreamReader, proc.stderr),
                errors,
            )
        ),
    )
    returncode = await proc.wait()

    completed = subprocess.CompletedProcess(
        args=args, returncode=returncode, stdout=stdout, stderr=stderr
    )
    if check:
        completed.check_returncode()
    return completed


def log_run(
    args: Sequence[StrOrBytesPath],
    logger: TwiggyLogger | StdLogger | Logger | None = None,
    stdout_loglevel: str | OutputCallbackType | None = None,
    stderr_loglevel: str | OutputCallbackType | None = "debug",
    check: bool = True,
    **kwargs,
) -> subprocess.CompletedProcess[str]:
    """
    Synchronous wrapper for the async_log_run function.
    This function runs a command in a subprocess and captures and logs its
    output.
    """
    return asyncio.run(
        async_log_run(args, logger, stdout_loglevel, stderr_loglevel, check, **kwargs)
    )


__all__ = (
    "async_log_run",
    "log_run",
    "CalledProcessError",
    "stdout_callback",
    "stderr_callback",
    "OutputCallbackType",
)



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Classes to encapsulate collection metadata from collection-meta.yaml
"""

from __future__ import annotations

import os
import typing as t

import pydantic as p
from antsibull_fileutils.yaml import load_yaml_file
from packaging.version import Version as PypiVer
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated, Self

if t.TYPE_CHECKING:
    from _typeshed import StrPath


def _convert_pypi_version(v: t.Any) -> t.Any:
    if isinstance(v, str):
        if not v:
            raise ValueError(f"must be a non-trivial string, got {v!r}")
        version = PypiVer(v)
    elif isinstance(v, PypiVer):
        version = v
    else:
        raise ValueError(f"must be a string or PypiVer object, got {v!r}")

    if len(version.release) != 3:
        raise ValueError(
            f"must be a version with three release numbers (e.g. 1.2.3, 2.3.4a1), got {v!r}"
        )
    return version


PydanticPypiVersion = Annotated[PypiVer, BeforeValidator(_convert_pypi_version)]


class RemovalUpdate(p.BaseModel):
    """
    Stores metadata about removal updates, like when a deprecation has been cancelled,
    the collection has been re-deprecated, when a removal has been undone, etc.
    """

    model_config = p.ConfigDict(arbitrary_types_allowed=True)

    # Exactly one of the following must be provided
    cancelled_version: t.Optional[PydanticPypiVersion] = None
    deprecated_version: t.Optional[PydanticPypiVersion] = None
    redeprecated_version: t.Optional[PydanticPypiVersion] = None
    removed_version: t.Optional[PydanticPypiVersion] = None
    readded_version: t.Optional[PydanticPypiVersion] = None

    # Overwrites the discussion link from BaseRemovalInformation if present
    discussion: t.Optional[str] = None

    # If deprecated_version or redeprecated_version: the reason because of which the
    # collection will be removed.
    reason: t.Optional[
        t.Literal[
            "deprecated",
            "considered-unmaintained",
            "renamed",
            "guidelines-violation",
            "other",
        ]
    ] = None

    # If reason is not provided, or if reason is 'other', an optional extra text appended
    # to the message.
    reason_text: t.Optional[str] = None

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _exactly_one_required(self) -> Self:
        count = sum(
            1 if x is not None else 0
            for x in (
                self.cancelled_version,
                self.deprecated_version,
                self.redeprecated_version,
                self.removed_version,
                self.readded_version,
            )
        )
        if count != 1:
            fields = (
                "cancelled_version",
                "deprecated_version",
                "redeprecated_version",
                "removed_version",
                "readded_version",
            )
            raise ValueError(f"Exactly one of {', '.join(fields)} must be specified")
        return self

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_reason(self) -> Self:
        if self.reason and not (self.deprecated_version or self.redeprecated_version):
            raise ValueError(
                "Reason can only be provided if 'deprecated_version'"
                " or 'redeprecated_version' is used"
            )
        return self

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_reason_text(self) -> Self:
        reasons_with_text = ("other", "guidelines-violation")
        if self.reason in reasons_with_text:
            if self.reason_text is None:
                raise ValueError(
                    f"Reason text must be provided if reason is '{self.reason}'"
                )
        elif self.reason:
            if self.reason_text is not None:
                raise ValueError(
                    f"Reason text must not be provided if reason is '{self.reason}'"
                )
        return self


class BaseRemovalInformation(p.BaseModel):
    """
    Stores metadata on why a collection was/will get removed.
    """

    model_config = p.ConfigDict(arbitrary_types_allowed=True)

    # The reason because of which the collection will be removed.
    reason: t.Literal[
        "deprecated",
        "considered-unmaintained",
        "renamed",
        "guidelines-violation",
        "other",
    ]
    reason_text: t.Optional[str] = None

    # The Ansible version in which the announcement was made. This is needed
    # for changelog generation.
    announce_version: t.Optional[PydanticPypiVersion] = None

    # In case reason=renamed, the new name of the collection.
    new_name: t.Optional[str] = None

    # The link to the discussion of the removal.
    discussion: t.Optional[p.HttpUrl] = None

    # In case reason=renamed, the major Ansible release in which the collection's
    # contents have been replaced by deprecated redirects.
    redirect_replacement_major_version: t.Optional[int] = None

    # Updates to the removal
    updates: list[RemovalUpdate] = []

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_reason_text(self) -> Self:
        reasons_with_text = ("other", "guidelines-violation")
        if self.reason in reasons_with_text:
            if self.reason_text is None:
                reasons = ", ".join(f"'{reason}'" for reason in reasons_with_text)
                raise ValueError(f"reason_text must be provided if reason is {reasons}")
        else:
            if self.reason_text is not None:
                reasons = ", ".join(f"'{reason}'" for reason in reasons_with_text)
                raise ValueError(
                    f"reason_text must not be provided if reason is not {reasons}"
                )
        return self

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_renamed_base(self) -> Self:
        if self.reason == "renamed":
            if self.new_name is None:
                raise ValueError("new_name must be provided if reason is 'renamed'")
        else:
            if self.new_name is not None:
                raise ValueError(
                    "new_name must not be provided if reason is not 'renamed'"
                )
            if self.redirect_replacement_major_version is not None:
                raise ValueError(
                    "redirect_replacement_major_version must not be provided"
                    " if reason is not 'renamed'"
                )
        return self


class RemovalInformation(BaseRemovalInformation):
    """
    Stores metadata on when and why a collection will get removed.
    """

    # The Ansible major version from which the collection will be removed.
    major_version: t.Union[int, t.Literal["TBD"]]

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_renamed(self) -> Self:
        if self.reason == "renamed":
            if (
                self.redirect_replacement_major_version is not None
                and self.major_version != "TBD"
                and self.redirect_replacement_major_version
                >= self.major_version  # pyre-ignore[58]
            ):
                raise ValueError(
                    "redirect_replacement_major_version must be smaller than major_version"
                )
        else:
            if self.major_version == "TBD":
                raise ValueError(
                    "major_version must not be TBD if reason is not 'renamed'"
                )
        return self

    def get_updates_including_indirect(self) -> list[RemovalUpdate]:
        prefix = []
        if self.announce_version:
            prefix.append(RemovalUpdate(deprecated_version=self.announce_version))
        return prefix + self.updates

    def is_deprecated(self) -> bool:
        result = True
        for update in self.get_updates_including_indirect():
            result = bool(
                update.deprecated_version
                or update.redeprecated_version
                or update.removed_version
            )
        return result


class RemovedRemovalInformation(BaseRemovalInformation):
    """
    Stores metadata on when and why a collection was removed.
    """

    # The exact version from which the collection has been removed.
    # This is needed for changelog generation.
    version: PydanticPypiVersion

    @p.model_validator(mode="after")  # pyre-ignore[56]
    def _check_reason_is_renamed(self) -> Self:
        if self.reason != "renamed":
            return self
        if (
            self.redirect_replacement_major_version is not None
            and self.redirect_replacement_major_version >= self.version.major
        ):
            raise ValueError(
                "redirect_replacement_major_version must be smaller than"
                " version's major version"
            )
        return self


class BaseCollectionMetadata(p.BaseModel):
    """
    Stores metadata about one collection.
    """

    # If the collection does not use changelogs/changelog.yaml, it can provide
    # a URL where the collection's changelog can be found.
    changelog_url: t.Optional[str] = p.Field(alias="changelog-url", default=None)

    # In case the collection is not located in the root of its repository, the
    # subdirectory in which the collection appears.
    collection_directory: t.Optional[str] = p.Field(
        alias="collection-directory", default=None
    )

    # The collection's repository.
    repository: t.Optional[str] = None

    # A regular expression to match the collection's version from a tag in the repository.
    tag_version_regex: t.Optional[str] = None

    # A list of maintainers. These should be usernames for the repository's
    # hosting environment.
    maintainers: list[str] = []


class CollectionMetadata(BaseCollectionMetadata):
    """
    Stores metadata about one collection.
    """

    model_config = p.ConfigDict(arbitrary_types_allowed=True)

    # Optional information that the collection will be removed from
    # a future Ansible release.
    removal: t.Optional[RemovalInformation] = None


class RemovedCollectionMetadata(BaseCollectionMetadata):
    """
    Stores metadata about a removed collection.
    """

    model_config = p.ConfigDict(arbitrary_types_allowed=True)

    # Information why the collection has been removed
    removal: RemovedRemovalInformation


class CollectionsMetadata(p.BaseModel):
    """
    Stores metadata about a set of collections.
    """

    # Metadata on the collections included in Ansible.
    collections: dict[str, CollectionMetadata]

    # Metadata on the collections removed from this major version of Ansible.
    removed_collections: dict[str, RemovedCollectionMetadata] = {}

    @staticmethod
    def load_from(deps_dir: StrPath | None) -> CollectionsMetadata:
        if deps_dir is None:
            return CollectionsMetadata(collections={})
        collection_meta_path = os.path.join(deps_dir, "collection-meta.yaml")
        if not os.path.exists(collection_meta_path):
            return CollectionsMetadata(collections={})
        data = load_yaml_file(collection_meta_path)
        return CollectionsMetadata.model_validate(data)

    def get_meta(self, collection_name: str) -> CollectionMetadata:
        result = self.collections.get(collection_name)
        if result is None:
            result = CollectionMetadata()
            self.collections[collection_name] = result
        return result



# Copyright (C) 2018 Julien Hartmann
# This program is distributed under the MIT license, a copy of which you should
# have receveived along with it. If not, see <https://opensource.org/licenses/MIT>.
#
# SPDX-License-Identifier: MIT
#
# https://github.com/solid-abstractions/python/tree/master/testing-aiohttp-clients

import asyncio
import collections
import inspect
import socket
import weakref

import aiohttp
import aiohttp.test_utils
import aiohttp.web
import pytest

# ----------------------------------------------------------------------------

_RedirectContext = collections.namedtuple("RedirectContext", "add_server session")


@pytest.fixture
async def http_redirect(ssl_certificate):
    """An HTTP ClientSession fixture that redirects requests to local test servers"""
    ssl_client_context = ssl_certificate.client_context()
    resolver = FakeResolver()
    connector = aiohttp.TCPConnector(
        resolver=resolver,
        ssl=ssl_client_context,
        use_dns_cache=False,
    )
    async with aiohttp.ClientSession(connector=connector) as session:
        yield _RedirectContext(add_server=resolver.add, session=session)


class FakeResolver:
    """aiohttp resolver that hijacks a set of uris

    :param servers: a mapping of remote host:port to redirect to local servers.
    :type servers: dict(tuple(str, int), int)
    """

    __slots__ = ("_servers",)

    def __init__(self, servers=None):
        self._servers = servers or {}

    def add(self, host, port, target):
        """Add an entry to the resolver"""
        self._servers[host, port] = target

    async def resolve(self, host, port=0, family=socket.AF_INET):
        """Resolve a host:port pair into a connectable address"""
        try:
            fake_port = self._servers[host, port]
        except KeyError:
            raise OSError("Fake DNS lookup failed: no fake server known for %s" % host)
        return [
            {
                "hostname": host,
                "host": "127.0.0.1",
                "port": fake_port,
                "family": socket.AF_INET,
                "proto": 0,
                "flags": socket.AI_NUMERICHOST,
            }
        ]


# ----------------------------------------------------------------------------


class CaseControlledTestServer(aiohttp.test_utils.RawTestServer):
    """Test server that relies on test case to supply responses and control timing"""

    def __init__(self, *, ssl=None, **kwargs):
        super().__init__(self._handle_request, **kwargs)
        self._ssl = ssl
        self._requests = asyncio.Queue()
        self._responses = {}

    async def start_server(self, **kwargs):
        kwargs.setdefault("ssl", self._ssl)
        await super().start_server(**kwargs)

    async def close(self):
        for future in self._responses.values():
            future.cancel()
        await super().close()

    async def _handle_request(self, request):
        self._responses[id(request)] = response = asyncio.Future()
        self._requests.put_nowait(request)
        try:
            return await response
        finally:
            del self._responses[id(request)]

    @property
    def awaiting_request_count(self):
        return self._requests.qsize()

    async def receive_request(self, *, timeout=None):
        """Wait until the test server receives a request

        :param float timeout: Bail out after that many seconds.
        :return: received request, not yet serviced.
        :rtype: aiohttp.web.BaseRequest
        :see: :meth:`send_response`
        """
        return await asyncio.wait_for(self._requests.get(), timeout=timeout)

    def send_response(self, request, *args, **kwargs):
        """Reply to a received request.

        :param request: the request to respond to.
        :param args: forwarded to :class:`aiohttp.web.Response`.
        :param kwargs: forwarded to :class:`aiohttp.web.Response`.
        """
        self._responses[id(request)].set_result(aiohttp.web.Response(*args, **kwargs))



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to work with Galaxy."""

from __future__ import annotations

import dataclasses
import os.path
import typing as t
from enum import Enum
from urllib.parse import urljoin

import aiofiles
import semantic_version as semver
from antsibull_fileutils.hashing import verify_hash
from antsibull_fileutils.io import copy_file

from . import app_context
from .utils.http import retry_get

# The type checker can handle finding aiohttp.client but flake8 cannot :-(
if t.TYPE_CHECKING:
    import aiohttp.client
    from _typeshed import StrPath


class NoSuchCollection(Exception):
    """Collection name does not map to a collection on Galaxy."""


class NoSuchVersion(Exception):
    """Version does not match with any versions of a collection on Galaxy."""


class DownloadFailure(Exception):
    """Failure downloading a collection from Galaxy."""


class DownloadResults(t.NamedTuple):
    """Results of downloading a collection."""

    #: :obj:`semantic_version.Version` of the exact version of the collection that was downloaded.
    version: semver.Version
    #: Location on the filesystem of the downloaded collection.
    download_path: str


class GalaxyVersion(Enum):
    V2 = 2
    V3 = 3


@dataclasses.dataclass
class GalaxyContext:
    server: str
    version: GalaxyVersion
    base_url: str

    @classmethod
    async def create(
        cls,
        aio_session: aiohttp.client.ClientSession,
        galaxy_server: t.Optional[str] = None,
    ) -> GalaxyContext:
        """
        Create a new Galaxy context.

        :arg aio_session: A ``aiohttp.client.ClientSession`` object.
        :kwarg galaxy_server: A Galaxy server URL. Defaults to ``lib_ctx.get().galaxy_url``.
        """
        if galaxy_server is None:
            galaxy_server = str(app_context.lib_ctx.get().galaxy_url)
        api_url = urljoin(galaxy_server, "api/")
        async with retry_get(
            aio_session, api_url, headers={"Accept": "application/json"}
        ) as response:
            galaxy_info = await response.json()
        available_versions: t.Mapping[str, str] = (
            galaxy_info.get("available_versions") or {}
        )
        if "v3" in available_versions:
            version = GalaxyVersion.V3
            base_url = urljoin(galaxy_server, "api/" + available_versions["v3"])
        elif "v2" in available_versions:
            version = GalaxyVersion.V2
            base_url = urljoin(galaxy_server, "api/" + available_versions["v2"])
        else:
            raise RuntimeError(
                f"Information retrieved from {api_url} seems to indicate"
                " neither Galaxy v2 API nor Galaxy v3 API"
            )
        return cls(galaxy_server, version, base_url)


class GalaxyClient:
    """Class for querying the Galaxy REST API."""

    def __init__(
        self,
        aio_session: aiohttp.client.ClientSession,
        *,
        context: GalaxyContext,
    ) -> None:
        """
        Create a GalaxyClient object to query the Galaxy Server.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :kwarg context: A ``GalaxyContext`` instance. Must be provided.
        """
        if context is None:
            raise ValueError(
                "The context parameter must be provided to the GalaxyClient constructor"
            )
        self.context = context
        self.aio_session = aio_session
        self.headers: dict[str, str] = {"Accept": "application/json"}
        self.params: dict[str, str] = {}
        if context:
            self._update_from_context(context)

    def _update_from_context(self, context: GalaxyContext) -> None:
        if context.version == GalaxyVersion.V2:
            self.params["format"] = "json"

    async def _get_galaxy_versions(
        self, context: GalaxyContext, versions_url: str, add_params: bool = True
    ) -> list[str]:
        """
        Retrieve the complete list of versions for a collection from a galaxy endpoint.

        This internal function retrieves versions for collections from a Galaxy endpoint.  If the
        information is paged, it continues to retrieve linked pages until all of the information has
        been returned.

        :arg context: the ``GalaxyContext`` to use.
        :arg version_url: url to the page to retrieve.
        :arg add_params: used internally during recursion. Do not specify when calling this.
        :returns: List of the all the versions of the collection.
        """
        if add_params:
            params = self.params.copy()
            if context.version == GalaxyVersion.V2:
                params["page_size"] = "100"
            else:
                params["limit"] = "50"
        else:
            params = None
        async with retry_get(
            self.aio_session,
            versions_url,
            params=params,
            headers=self.headers,
            acceptable_error_codes=[404],
        ) as response:
            if response.status == 404:
                raise NoSuchCollection(f"No collection found at: {versions_url}")
            collection_info = await response.json()

        versions = []
        if context.version == GalaxyVersion.V2:
            results = collection_info["results"]
            next_link = collection_info["next"]
        else:
            if "data" in collection_info:
                # Apparently 'data' isn't always used...
                results = collection_info["data"]
            else:
                results = collection_info["results"]
            next_link = collection_info["links"]["next"]
            add_params = False
        for version_record in results:
            versions.append(version_record["version"])
        if next_link:
            if next_link.startswith("/"):
                next_link = urljoin(context.server, next_link)
            versions.extend(
                await self._get_galaxy_versions(context, next_link, add_params)
            )

        return versions

    async def get_versions(self, collection: str) -> list[str]:
        """
        Retrieve all versions of a collection on Galaxy.

        :arg collection: Name of the collection to get version info for.
        :returns: List of all the versions of this collection on galaxy.
        """
        context = self.context

        collection = collection.replace(".", "/")
        galaxy_url = urljoin(context.base_url, f"collections/{collection}/versions/")
        retval = await self._get_galaxy_versions(context, galaxy_url)
        return retval

    async def get_info(self, collection: str) -> dict[str, t.Any]:
        """
        Retrieve information about the collection on Galaxy.

        :arg collection: Namespace.collection to retrieve information about.
        :returns: Dictionary of information about the collection.

        Please see the Galaxy v2 and v3 REST API documentation for information on the
        structure of the returned data.

        .. seealso::
            An example return value from the
            `Galaxy v2 REST API
            <https://galaxy.ansible.com/api/v2/collections/community/general/>`_
            and the `Galaxy v3 REST API
            <https://beta-galaxy.ansible.com/api/v3/plugin/ansible/content/published/collections/index/community/general/>`_
        """
        context = self.context

        collection = collection.replace(".", "/")
        galaxy_url = urljoin(context.base_url, f"collections/{collection}/")

        async with retry_get(
            self.aio_session,
            galaxy_url,
            params=self.params,
            headers=self.headers,
            acceptable_error_codes=[404],
        ) as response:
            if response.status == 404:
                raise NoSuchCollection(f"No collection found at: {galaxy_url}")
            collection_info = await response.json()

        return collection_info

    async def get_release_info(
        self, collection: str, version: str | semver.Version
    ) -> dict[str, t.Any]:
        """
        Retrive information about a specific version of a collection.

        :arg collection: Namespace.collection string naming the collection.
        :arg version: Version of the collection.
        :returns: Dictionary of information about the release.

        Please see the Galaxy v2 and v3 REST API documentation for information on the
        structure of the returned data.

        .. seealso::
            An example return value from the
            `Galaxy v2 REST API
            <https://galaxy.ansible.com/api/v2/collections/community/general/versions/0.1.1>`_
            and the `Galaxy v3 REST API
            <https://beta-galaxy.ansible.com/api/v3/plugin/ansible/content/published/collections/index/community/general/versions/0.1.1/>`_
        """
        context = self.context

        collection = collection.replace(".", "/")
        galaxy_url = urljoin(
            context.base_url, f"collections/{collection}/versions/{version}/"
        )

        async with retry_get(
            self.aio_session,
            galaxy_url,
            params=self.params,
            headers=self.headers,
            acceptable_error_codes=[404],
        ) as response:
            if response.status == 404:
                raise NoSuchCollection(f"No collection found at: {galaxy_url}")
            collection_info = await response.json()

        return collection_info

    async def get_latest_matching_version(
        self, collection: str, version_spec: str, pre: bool = False
    ) -> semver.Version:
        """
        Get the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :kwarg pre: If True, allow prereleases (versions which have the form X.Y.Z.SOMETHING).
            This is **not** for excluding 0.Y.Z versions.  non-pre-releases are still
            preferred over pre-releases (for instance, with version_spec='>2.0.0-a1,<3.0.0'
            and pre=True, if the available versions are 2.0.0-a1 and 2.0.0-a2, then 2.0.0-a2
            will be returned.  If the available versions are 2.0.0 and 2.1.0-b2, 2.0.0 will be
            returned since non-pre-releases are preferred.  The default is False
        :returns: :obj:`semantic_version.Version` of the latest collection version that satisfied
            the specification.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`

        .. versionchanged:: 0.37.0
            Giving True to the ``pre`` parameter now means that prereleases will be
            *allowed* but stable releases will still be *preferred*.  Previously, the
            latest release, whether stable or prerelease was returned when pre was True.
        """
        versions = await self.get_versions(collection)
        sem_versions = [semver.Version(v) for v in versions]
        sem_versions.sort(reverse=True)

        spec = semver.SimpleSpec(version_spec)
        prereleases = []
        for version in (v for v in sem_versions if v in spec):
            # If this is a pre-release, first check if there's a non-pre-release that
            # will satisfy the version_spec.
            if version.prerelease:
                prereleases.append(version)
                continue
            return version

        # We did not find a stable version that satisies the version_spec.  If we
        # allow prereleases, return the latest of those here.
        if pre and prereleases:
            return prereleases[0]

        # No matching versions were found
        raise NoSuchVersion(
            f"{version_spec} did not match with any version of {collection}."
        )


class CollectionDownloader(GalaxyClient):
    """Manage downloading collections from Galaxy."""

    def __init__(
        self,
        aio_session: aiohttp.client.ClientSession,
        download_dir: StrPath,
        *,
        collection_cache: str | None = None,
        context: GalaxyContext,
        trust_collection_cache: t.Optional[bool] = None,
    ) -> None:
        """
        Create an object to download collections from galaxy.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :arg download_dir: Directory to download into.
        :kwarg context: A ``GalaxyContext`` instance. Must be provided.
        :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
            These tarballs will be used instead of downloading new tarballs provided that the
            versions match the criteria (latest compatible version known to galaxy).
            Defaults to ``lib_ctx.get().collection_cache``.
        :kwarg trust_collection_cache: If set to ``True``, will assume that if the collection
            cache contains an artifact, it is the current one available on the Galaxy server.
            This avoids making a request to the Galaxy server to figure out the artifact's
            checksum and comparing it before trusting the cached artifact.
        """
        super().__init__(aio_session, context=context)
        self.download_dir = download_dir
        lib_ctx = app_context.lib_ctx.get()
        if collection_cache is None:
            collection_cache = lib_ctx.collection_cache
        self.collection_cache: t.Final[str | None] = collection_cache
        if trust_collection_cache is None:
            trust_collection_cache = lib_ctx.trust_collection_cache
        self.trust_collection_cache: t.Final[bool] = trust_collection_cache

    async def download(
        self,
        collection: str,
        version: str | semver.Version,
    ) -> str:
        """
        Download a collection.

        Downloads the collection at the specified version.

        :arg collection: Namespace.collection identifying the collection.
        :arg version: Version of the collection to download.
        :returns: The full path to the downloaded collection.
        """
        namespace, name = collection.split(".", 1)
        filename = f"{namespace}-{name}-{version}.tar.gz"
        download_filename = os.path.join(self.download_dir, filename)
        lib_ctx = app_context.lib_ctx.get()

        if self.collection_cache and self.trust_collection_cache:
            cached_copy = os.path.join(self.collection_cache, filename)
            if os.path.isfile(cached_copy):
                await copy_file(
                    cached_copy,
                    download_filename,
                    check_content=False,
                    file_check_content=lib_ctx.file_check_content,
                    chunksize=lib_ctx.chunksize,
                )
                return download_filename

        release_info = await self.get_release_info(f"{namespace}/{name}", version)
        release_url = release_info["download_url"]

        sha256sum = release_info["artifact"]["sha256"]

        if self.collection_cache:
            cached_copy = os.path.join(self.collection_cache, filename)
            if os.path.isfile(cached_copy):
                lib_ctx = app_context.lib_ctx.get()
                if await verify_hash(
                    cached_copy, sha256sum, chunksize=lib_ctx.chunksize
                ):
                    await copy_file(
                        cached_copy,
                        download_filename,
                        check_content=False,
                        file_check_content=lib_ctx.file_check_content,
                        chunksize=lib_ctx.chunksize,
                    )
                    return download_filename

        async with retry_get(
            self.aio_session, release_url, acceptable_error_codes=[404]
        ) as response:
            if response.status == 404:
                raise NoSuchCollection(f"No collection found at: {release_url}")

            async with aiofiles.open(download_filename, "wb") as f:
                lib_ctx = app_context.lib_ctx.get()
                while chunk := await response.content.read(lib_ctx.chunksize):
                    await f.write(chunk)

        # Verify the download
        if not await verify_hash(
            download_filename, sha256sum, chunksize=lib_ctx.chunksize
        ):
            raise DownloadFailure(
                f"{release_url} failed to download correctly."
                f" Expected checksum: {sha256sum}"
            )

        # Copy downloaded collection into cache
        if self.collection_cache:
            cached_copy = os.path.join(self.collection_cache, filename)
            await copy_file(
                download_filename,
                cached_copy,
                check_content=False,
                file_check_content=lib_ctx.file_check_content,
                chunksize=lib_ctx.chunksize,
            )

        return download_filename

    async def download_latest_matching(
        self, collection: str, version_spec: str
    ) -> DownloadResults:
        """
        Download the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :returns: :obj:`DownloadResults` with version and download path for the collection we
            downloaded.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`
        """
        version = await self.get_latest_matching_version(collection, version_spec)
        download_path = await self.download(collection, version)
        return DownloadResults(version=version, download_path=download_path)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project
"""Schemas for logging config."""

import os.path
import typing as t
from collections.abc import Callable, Mapping, MutableMapping, MutableSequence, Sequence

import pydantic as p
import twiggy.formats  # type: ignore[import]
import twiggy.outputs  # type: ignore[import]

#: Valid choices for a logging level field
LEVEL_CHOICES_F = p.Field(
    ..., pattern="^(CRITICAL|ERROR|WARNING|NOTICE|INFO|DEBUG|DISABLED)$"
)

#: Valid choice of the logging version field
VERSION_CHOICES_F = p.Field(..., pattern=r"1\.0")


#
# Configuration file schema
#


class BaseModel(p.BaseModel):
    model_config = p.ConfigDict(frozen=True, extra="forbid", validate_default=True)


class LogFiltersModel(BaseModel):
    filter: t.Union[str, Callable]
    args: Sequence[t.Any] = []
    kwargs: Mapping[str, t.Any] = {}


class LogEmitterModel(BaseModel):
    output_name: str
    level: str = LEVEL_CHOICES_F
    filters: list[LogFiltersModel] = []


class LogOutputModel(BaseModel):
    output: t.Union[str, Callable]
    args: Sequence[t.Any] = []
    format: t.Union[str, Callable] = twiggy.formats.line_format
    kwargs: Mapping[str, t.Any] = {}

    @p.field_validator("args")
    # pylint:disable=no-self-argument
    def expand_home_dir_args(
        cls, args_field: MutableSequence, info: p.ValidationInfo
    ) -> MutableSequence:
        """Expand tilde in the arguments of specific outputs."""
        values = info.data
        if values["output"] in ("twiggy.outputs.FileOutput", twiggy.outputs.FileOutput):
            if args_field:
                args_field[0] = os.path.expanduser(args_field[0])
        return args_field

    @p.field_validator("kwargs")
    # pylint:disable=no-self-argument
    def expand_home_dir_kwargs(
        cls, kwargs_field: MutableMapping, info: p.ValidationInfo
    ) -> MutableMapping:
        """Expand tilde in the keyword arguments of specific outputs."""
        values = info.data
        if values["output"] in ("twiggy.outputs.FileOutput", twiggy.outputs.FileOutput):
            if "name" in kwargs_field:
                kwargs_field["name"] = os.path.expanduser(kwargs_field["name"])
        return kwargs_field


class LoggingModel(BaseModel):
    emitters: t.Optional[dict[str, LogEmitterModel]] = {}
    incremental: bool = False
    outputs: t.Optional[dict[str, LogOutputModel]] = {}
    version: str = VERSION_CHOICES_F


#: Default logging configuration
DEFAULT_LOGGING_CONFIG = LoggingModel.model_validate(
    {
        "version": "1.0",
        "outputs": {
            "logfile": {
                "output": "twiggy.outputs.FileOutput",
                "args": ["~/antsibull.log"],
            },
            "stderr": {
                "output": "twiggy.outputs.StreamOutput",
                "format": "twiggy.formats.shell_format",
            },
        },
        "emitters": {
            "all": {"level": "INFO", "output_name": "logfile", "filters": []},
            "problems": {"level": "WARNING", "output_name": "stderr", "filters": []},
        },
    }
)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Compat for older versions of Python."""

from __future__ import annotations

import argparse
import sys

BooleanOptionalAction: type[argparse.BooleanOptionalAction]

if sys.version_info < (3, 9, 11) or (
    sys.version_info >= (3, 10, 0) and sys.version_info < (3, 10, 3)
):
    # https://bugs.python.org/issue46080 was fixed in Python 3.11.0 alpha 5
    # (https://docs.python.org/3/whatsnew/changelog.html#python-3-11-0-alpha-5)
    # and backported to Python 3.10.3
    # (https://docs.python.org/3.10/whatsnew/changelog.html#python-3-10-3-final)
    # and Python 3.9.11
    # (https://docs.python.org/3.9/whatsnew/changelog.html#python-3-9-11-final).
    # Versions before these have to use the vendored version so that users can actually
    # use `--help` for affected subcommands.
    from .vendored._argparse_booleanoptionalaction import (  # type: ignore
        BooleanOptionalAction,
    )
else:
    BooleanOptionalAction = argparse.BooleanOptionalAction


__all__ = ("BooleanOptionalAction",)



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Classes to encapsulate collection metadata from collection-meta.yaml
"""

from __future__ import annotations

import os
import typing as t
from collections.abc import Collection

import pydantic as p
from antsibull_fileutils.yaml import load_yaml_file
from packaging.version import Version as PypiVer

from .pydantic import forbid_extras, get_formatted_error_messages
from .schemas.collection_meta import (
    BaseRemovalInformation,
    CollectionMetadata,
    CollectionsMetadata,
    RemovalInformation,
    RemovalUpdate,
    RemovedCollectionMetadata,
    RemovedRemovalInformation,
)

if t.TYPE_CHECKING:
    from _typeshed import StrPath


class _Validator:
    """
    Validate a CollectionsMetadata object.

    Validation error messages are added to the ``errors`` attribute.
    """

    errors: list[str]

    def __init__(self, *, all_collections: list[str], major_release: int):
        self.errors = []
        self.all_collections = all_collections
        self.major_release = major_release

    def _update_state_value(
        self,
        state: str | None,
        accepted_states: list[str | None],
        prefix: str,
        index: int,
        field_name: str,
    ) -> str:
        if state not in accepted_states:
            if state is not None:
                self.errors.append(
                    f"{prefix}[{index}] -> {field_name}: Unexpected update after {state}"
                )
            else:
                self.errors.append(
                    f"{prefix}[{index}] -> {field_name}: Unexpected first update"
                )
        return field_name

    def _update_state(
        self, state: str | None, index: int, update: RemovalUpdate, prefix: str
    ) -> tuple[str | None, PypiVer | None, str]:
        if update.cancelled_version:
            state = self._update_state_value(
                state,
                [None, "deprecated_version", "redeprecated_version"],
                prefix,
                index,
                "cancelled_version",
            )
            return state, update.cancelled_version, "cancelled_version"
        if update.deprecated_version:
            state = self._update_state_value(
                state, [None], prefix, index, "deprecated_version"
            )
            return state, update.deprecated_version, "deprecated_version"
        if update.redeprecated_version:
            state = self._update_state_value(
                state,
                ["readded_version", "cancelled_version"],
                prefix,
                index,
                "redeprecated_version",
            )
            return state, update.redeprecated_version, "redeprecated_version"
        if update.removed_version:
            state = self._update_state_value(
                state, [None], prefix, index, "removed_version"
            )
            return state, update.removed_version, "removed_version"
        if update.readded_version:
            state = self._update_state_value(
                state, ["removed_version"], prefix, index, "readded_version"
            )
            return state, update.readded_version, "readded_version"
        # The following lines should never be reached:
        self.errors.append(f"{prefix}[{index}]: Internal error")  # pragma: no cover
        return state, None, ""  # pragma: no cover

    def _validate_removal_updates(
        self,
        removal: BaseRemovalInformation,
        indirect_updates: list[RemovalUpdate],
        prefix: str,
    ) -> None:
        prefix += " -> updates"
        state = None
        for update in indirect_updates:
            state, _, __ = self._update_state(state, -1, update, prefix)
        last_version = None
        for index, update in enumerate(removal.updates):
            state, version, field_name = self._update_state(
                state, index, update, prefix
            )
            if version is None:
                # The following line should never be reached:
                pass  # pragma: no cover
            elif version.major != self.major_release:
                self.errors.append(
                    f"{prefix}[{index}] -> {field_name}: Version's major version {version.major}"
                    f" must be the current major version {self.major_release}"
                )
            elif last_version is not None and version <= last_version:
                self.errors.append(
                    f"{prefix}[{index}] -> {field_name}: Version {version}"
                    f" must be after the previous update's version {last_version}"
                )
            last_version = version

    def _validate_removal_base(
        self, collection: str, removal: BaseRemovalInformation, prefix: str
    ) -> None:
        if removal.reason == "renamed" and removal.new_name == collection:
            self.errors.append(f"{prefix} new_name: Must not be the collection's name")

    def _validate_removal(
        self, collection: str, removal: RemovalInformation, prefix: str
    ) -> None:
        if (
            removal.major_version != "TBD"
            and removal.major_version <= self.major_release  # pyre-ignore[58]
        ):
            is_ok = removal.major_version == self.major_release and any(
                update.removed_version for update in removal.updates
            )
            if not is_ok:
                self.errors.append(
                    f"{prefix} major_version: Removal major version {removal.major_version} must"
                    f" be larger than current major version {self.major_release}"
                )

        if (
            removal.announce_version is not None
            and removal.announce_version.major > self.major_release
        ):
            self.errors.append(
                f"{prefix} announce_version: Major version of {removal.announce_version}"
                f" must not be larger than the current major version {self.major_release}"
            )

        self._validate_removal_base(collection, removal, prefix)

        indirect_updates = []
        if removal.announce_version is not None:
            indirect_updates.append(
                RemovalUpdate(
                    deprecated_version=removal.announce_version,
                    reason=removal.reason,
                    reason_text=removal.reason_text,
                )
            )
        self._validate_removal_updates(removal, indirect_updates, prefix)

    def _validate_collection(
        self, collection: str, meta: CollectionMetadata, prefix: str
    ) -> None:
        if meta.repository is None:
            self.errors.append(f"{prefix} repository: Required field not provided")

        if meta.removal:
            self._validate_removal(collection, meta.removal, f"{prefix} removal ->")

    def _validate_removal_for_removed(
        self, collection: str, removal: RemovedRemovalInformation, prefix: str
    ) -> None:
        if removal.version.major > self.major_release:
            self.errors.append(
                f"{prefix} version: Major version of removal version {removal.version} must"
                f" not be larger than current major version {self.major_release}"
            )

        if (
            removal.announce_version is not None
            and removal.announce_version.major >= self.major_release
        ):
            self.errors.append(
                f"{prefix} announce_version: Major version of {removal.announce_version}"
                f" must be less than the current major version {self.major_release}"
            )

        self._validate_removal_base(collection, removal, prefix)

    def _validate_removed_collection(
        self, collection: str, meta: RemovedCollectionMetadata, prefix: str
    ) -> None:
        if meta.repository is None:
            self.errors.append(f"{prefix} repository: Required field not provided")

        self._validate_removal_for_removed(
            collection, meta.removal, f"{prefix} removal ->"
        )

    def _validate_order(self, collection: Collection, what: str) -> None:
        # Check order
        sorted_list = sorted(collection)
        raw_list = list(collection)
        for raw_entry, sorted_entry in zip(raw_list, sorted_list):
            if raw_entry != sorted_entry:
                self.errors.append(
                    f"{what} must be sorted; "
                    f"{sorted_entry!r} must come before {raw_entry}"
                )
                break

    def _validate_collections(self, data: CollectionsMetadata) -> None:
        # Check order
        self._validate_order(data.collections, "The collection list")

        # Validate collection data
        remaining_collections = set(self.all_collections)
        for collection, meta in data.collections.items():
            if collection not in remaining_collections:
                self.errors.append(
                    f"collections -> {collection}: Collection not in ansible.in"
                )
            else:
                remaining_collections.remove(collection)
            self._validate_collection(
                collection, meta, f"collections -> {collection} ->"
            )

        # Complain about remaining collections
        for collection in sorted(remaining_collections):
            self.errors.append(f"collections: No metadata present for {collection}")

    def _validate_removed_collections(self, data: CollectionsMetadata) -> None:
        # Check order
        self._validate_order(data.removed_collections, "The removed collection list")

        # Validate removed collection data
        for collection, removed_meta in data.removed_collections.items():
            if collection in self.all_collections:
                self.errors.append(
                    f"removed_collections -> {collection}: Collection in ansible.in"
                )
            self._validate_removed_collection(
                collection, removed_meta, f"removed_collections -> {collection} ->"
            )

    def validate(self, data: CollectionsMetadata) -> None:
        self._validate_collections(data)
        self._validate_removed_collections(data)


def lint_collection_meta(
    *, collection_meta_path: StrPath, major_release: int, all_collections: list[str]
) -> list[str]:
    """Lint collection-meta.yaml."""
    if not os.path.exists(collection_meta_path):
        return [f"Cannot find {collection_meta_path}"]

    try:
        data = load_yaml_file(collection_meta_path)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [f"Error while parsing YAML file: {exc}"]

    validator = _Validator(
        all_collections=all_collections,
        major_release=major_release,
    )

    forbid_extras(CollectionsMetadata)

    try:
        parsed_data = CollectionsMetadata.model_validate(data)
        validator.validate(parsed_data)
    except p.ValidationError as exc:
        validator.errors.extend(get_formatted_error_messages(exc))

    return sorted(validator.errors)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project

"""
Shared code used by tools for building the Ansible distribution
"""

from __future__ import annotations

__version__ = "3.5.0"

__all__ = ("__version__",)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import os
from pathlib import Path

import pytest

from antsibull_core.collection_meta import lint_collection_meta
from antsibull_core.schemas.collection_meta import (
    CollectionMetadata,
    CollectionsMetadata,
    RemovalInformation,
    RemovalUpdate,
)

LINT_COLLECTION_META_DATA = [
    (
        5,
        r"""[ invalid yaml
""",
        [],
        [
            "Error while parsing YAML file: while parsing a flow sequence\n"
            '  in "{filename}", line 1, column 1\n'
            "did not find expected ',' or ']'\n"
            '  in "{filename}", line 2, column 1',
        ],
    ),
    (
        10,
        r"""---
collections: {}
""",
        [],
        [],
    ),
    (
        10,
        r"""---
collections:
  baz.bam:
    repository: https://github.com/ansible-collections/collection_template
  foo.bar:
    repository: https://github.com/ansible-collections/collection_template
""",
        [
            "foo.bar",
            "baz.bam",
        ],
        [],
    ),
    (
        9,
        r"""---
collections:
  bad.bar1:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 7
      reason: deprecated
      announce_version: 10.1.0
      updates:
        - removed_version: 9.2.0
        - cancelled_version: 9.3.0
        - cancelled_version: 9.4.0
        - redeprecated_version: 9.4.0
        - removed_version: 9.3.0
        - readded_version: 10.6.0
  bad.bar2:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: renamed
      new_name: bad.bar2
      announce_version: 9.3.0
      redirect_replacement_major_version: 9
  foo.bar: {}
  baz.bam: {}
  correct.foo1:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  correct.foo10:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 9
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
      updates:
        - removed_version: 9.0.0a1
        - readded_version: 9.1.0
  correct.foo11:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 9
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  correct.foo2:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: deprecated
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  correct.foo3:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 10
      reason: renamed
      new_name: namespace.name
      updates:
        - readded_version: 10.1.0
  correct.foo4:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: TBD
      reason: renamed
      new_name: namespace.name
      announce_version: 9.3.0
  correct.foo5:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: renamed
      new_name: namespace.name
      announce_version: 9.3.0
      redirect_replacement_major_version: 10
  correct.foo6:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: TBD
      reason: renamed
      new_name: namespace.name
      announce_version: 9.3.0
      redirect_replacement_major_version: 12
  correct.foo7:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: guidelines-violation
      reason_text: There was a violation.
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  correct.foo8:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 12
      reason: guidelines-violation
      reason_text: Refused to run CI.
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  correct.foo9:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: other
      reason_text: The collection wasn't cow friendly, so the Steering Committee decided to kick it out.
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
removed_collections:
  bad.baz2:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      version: 10.2.1
      reason: renamed
      new_name: bad.bar2
      announce_version: 9.3.0
      redirect_replacement_major_version: 7
  bad.baz1:
    removal:
      version: 9.1.0
      reason: deprecated
      announce_version: 7.1.0
  foo.bar:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      version: 9.0.0
      reason: deprecated
      announce_version: 7.1.0
""",
        [
            "foo.bar",
            "correct.foo1",
            "correct.foo10",
            "correct.foo11",
            "correct.foo2",
            "correct.foo3",
            "correct.foo4",
            "correct.foo5",
            "correct.foo6",
            "correct.foo7",
            "correct.foo8",
            "correct.foo9",
            "not.there",
        ],
        [
            "The collection list must be sorted; 'baz.bam' must come before foo.bar",
            "The removed collection list must be sorted; 'bad.baz1' must come before bad.baz2",
            "collections -> bad.bar1 -> removal -> -> updates[0] -> removed_version: Unexpected update after deprecated_version",
            "collections -> bad.bar1 -> removal -> -> updates[1] -> cancelled_version: Unexpected update after removed_version",
            "collections -> bad.bar1 -> removal -> -> updates[2] -> cancelled_version: Unexpected update after cancelled_version",
            "collections -> bad.bar1 -> removal -> -> updates[3] -> redeprecated_version: Version 9.4.0 must be after the previous update's version 9.4.0",
            "collections -> bad.bar1 -> removal -> -> updates[4] -> removed_version: Unexpected update after redeprecated_version",
            "collections -> bad.bar1 -> removal -> -> updates[4] -> removed_version: Version 9.3.0 must be after the previous update's version 9.4.0",
            "collections -> bad.bar1 -> removal -> -> updates[5] -> readded_version: Version's major version 10 must be the current major version 9",
            "collections -> bad.bar1 -> removal -> announce_version: Major version of 10.1.0 must not be larger than the current major version 9",
            "collections -> bad.bar1 -> removal -> major_version: Removal major version 7 must be larger than current major version 9",
            "collections -> bad.bar1: Collection not in ansible.in",
            "collections -> bad.bar2 -> removal -> new_name: Must not be the collection's name",
            "collections -> bad.bar2: Collection not in ansible.in",
            "collections -> baz.bam -> repository: Required field not provided",
            "collections -> baz.bam: Collection not in ansible.in",
            "collections -> correct.foo10 -> removal -> -> updates[0] -> removed_version: Unexpected update after deprecated_version",
            "collections -> correct.foo11 -> removal -> major_version: Removal major version 9 must be larger than current major version 9",
            "collections -> correct.foo3 -> removal -> -> updates[0] -> readded_version: Unexpected first update",
            "collections -> correct.foo3 -> removal -> -> updates[0] -> readded_version: Version's major version 10 must be the current major version 9",
            "collections -> foo.bar -> repository: Required field not provided",
            "collections: No metadata present for not.there",
            "removed_collections -> bad.baz1 -> repository: Required field not provided",
            "removed_collections -> bad.baz2 -> removal -> announce_version: Major version of 9.3.0 must be less than the current major version 9",
            "removed_collections -> bad.baz2 -> removal -> version: Major version of removal version 10.2.1 must not be larger than current major version 9",
            "removed_collections -> foo.bar: Collection in ansible.in",
        ],
    ),
    (
        9,
        r"""---
collections:
  bad.foo0:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: renamed
      announce_version: 9.3.0
  bad.foo1:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
      announce_version: 42
  bad.foo2:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: deprecated
      discussion: https://forum.ansible.com/...
      announce_version: "9.3"
  bad.foo3:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 10
      reason: renamed
      new_name: namespace.name
      announce_version: ''
  bad.foo4:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: TBD
      reason: deprecated
      announce_version: 9.3.0
  bad.foo5:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: renamed
      new_name: namespace.name
      announce_version: 9.3.0
      redirect_replacement_major_version: 11
  bad.foo6:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: TBD
      reason: renamed
      reason_text: There was a violation.
      new_name: namespace.name
      announce_version: 9.3.0
      redirect_replacement_major_version: 12
  bad.foo7:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: guidelines-violation
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  bad.foo8:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 12
      reason: guidelines-violation
      reason_text: Refused to run CI.
      new_name: namespace.name
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
  bad.foo9:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: other
      reason_text: The collection wasn't cow friendly, so the Steering Committee decided to kick it out.
      discussion: https://forum.ansible.com/...
      announce_version: 9.3.0
      redirect_replacement_major_version: 12
  bad.foo10:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: considered-unmaintained
      reason_text: Foo!
      reason_other: This shouldn't be there!
      discussion: https://forum.ansible.com/...
  bad.foo11:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
  bad.foo12:
    repository: https://github.com/ansible-collections/collection_template
    extra_stuff: bam
  bad.foo13:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: foo
      reason: considered-unmaintained
      discussion: https://forum.ansible.com/...
  bad.foo14:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: foo
  bad.foo15:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      major_version: 11
      reason: considered-unmaintained
      updates:
        - {}
        - foo: bar
        - cancelled_version: 1.2.3
          extra: bad
        - cancelled_version: []
        - cancelled_version: 1.2.3
          deprecated_version: 1.2.3
          redeprecated_version: 1.2.3
          removed_version: 1.2.3
          readded_version: 1.2.3
        - reason_text: foo
        - readded_version: 1.2.3
          reason: deprecated
        - deprecated_version: 1.2.3
          reason: deprecated
          reason_text: bla
        - deprecated_version: 1.2.3
          reason: other
removed_collections:
  bad.foo1:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      version: TBD
      reason: deprecated
  bad.foo2:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      reason: deprecated
  bad.foo3:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      version: 9.0.0
      reason: renamed
      new_name: bad.foo3_new
      redirect_replacement_major_version: 12
  bad.foo4:
    repository: https://github.com/ansible-collections/collection_template
    removal:
      version: 9.0.0
      reason: renamed
      new_name: bad.foo3_new
      updates: {}
extra_stuff: baz
""",
        [],
        [
            "collections -> bad.foo0 -> removal: Value error, new_name must be provided if reason is 'renamed'",
            "collections -> bad.foo1 -> removal -> announce_version: Value error, must be a string or PypiVer object, got 42",
            "collections -> bad.foo10 -> removal -> reason_other: Extra inputs are not permitted",
            "collections -> bad.foo12 -> extra_stuff: Extra inputs are not permitted",
            "collections -> bad.foo13 -> removal -> major_version -> int: Input should be a valid integer, unable to parse string as an integer",
            "collections -> bad.foo13 -> removal -> major_version -> literal['TBD']: Input should be 'TBD'",
            "collections -> bad.foo14 -> removal -> reason: Input should be 'deprecated', 'considered-unmaintained', 'renamed', 'guidelines-violation' or 'other'",
            "collections -> bad.foo15 -> removal -> updates -> 0: Value error, Exactly one of cancelled_version, deprecated_version, redeprecated_version, removed_version, readded_version must be specified",
            "collections -> bad.foo15 -> removal -> updates -> 1 -> foo: Extra inputs are not permitted",
            "collections -> bad.foo15 -> removal -> updates -> 2 -> extra: Extra inputs are not permitted",
            "collections -> bad.foo15 -> removal -> updates -> 3 -> cancelled_version: Value error, must be a string or PypiVer object, got []",
            "collections -> bad.foo15 -> removal -> updates -> 4: Value error, Exactly one of cancelled_version, deprecated_version, redeprecated_version, removed_version, readded_version must be specified",
            "collections -> bad.foo15 -> removal -> updates -> 5: Value error, Exactly one of cancelled_version, deprecated_version, redeprecated_version, removed_version, readded_version must be specified",
            "collections -> bad.foo15 -> removal -> updates -> 6: Value error, Reason can only be provided if 'deprecated_version' or 'redeprecated_version' is used",
            "collections -> bad.foo15 -> removal -> updates -> 7: Value error, Reason text must not be provided if reason is 'deprecated'",
            "collections -> bad.foo15 -> removal -> updates -> 8: Value error, Reason text must be provided if reason is 'other'",
            "collections -> bad.foo2 -> removal -> announce_version: Value error, must be a version with three release numbers (e.g. 1.2.3, 2.3.4a1), got '9.3'",
            "collections -> bad.foo3 -> removal -> announce_version: Value error, must be a non-trivial string, got ''",
            "collections -> bad.foo4 -> removal: Value error, major_version must not be TBD if reason is not 'renamed'",
            "collections -> bad.foo5 -> removal: Value error, redirect_replacement_major_version must be smaller than major_version",
            "collections -> bad.foo6 -> removal: Value error, reason_text must not be provided if reason is not 'other', 'guidelines-violation'",
            "collections -> bad.foo7 -> removal: Value error, reason_text must be provided if reason is 'other', 'guidelines-violation'",
            "collections -> bad.foo8 -> removal: Value error, new_name must not be provided if reason is not 'renamed'",
            "collections -> bad.foo9 -> removal: Value error, redirect_replacement_major_version must not be provided if reason is not 'renamed'",
            "extra_stuff: Extra inputs are not permitted",
            "removed_collections -> bad.foo1 -> removal -> version: Value error, Invalid version: 'TBD'",
            "removed_collections -> bad.foo2 -> removal -> version: Field required",
            "removed_collections -> bad.foo3 -> removal: Value error, redirect_replacement_major_version must be smaller than version's major version",
            "removed_collections -> bad.foo4 -> removal -> updates: Input should be a valid list",
        ],
    ),
]


@pytest.mark.parametrize(
    "major_release, collection_metadata, all_collections, expected_errors",
    LINT_COLLECTION_META_DATA,
)
def test_lint_collection_meta(
    major_release: int,
    collection_metadata: str,
    all_collections: list[str],
    expected_errors: list[str],
    tmp_path: Path,
):
    filename = tmp_path / "collection-meta.yaml"
    filename.write_text(collection_metadata)
    errors = lint_collection_meta(
        collection_meta_path=filename,
        major_release=major_release,
        all_collections=all_collections,
    )
    assert errors == [
        error.replace("{filename}", str(filename)) for error in expected_errors
    ]


def test_lint_collection_meta_not_existing(tmp_path: Path):
    filename = tmp_path / "collection-meta.yaml"
    errors = lint_collection_meta(
        collection_meta_path=filename, major_release=5, all_collections=["foo.bar"]
    )
    assert errors == [f"Cannot find {filename}"]

    filename.mkdir()
    errors = lint_collection_meta(
        collection_meta_path=filename, major_release=5, all_collections=["foo.bar"]
    )
    assert errors == [
        f"Error while parsing YAML file: [Errno 21] Is a directory: {str(filename)!r}"
    ]


def test_collections_metadata_methods(tmp_path: Path):
    assert CollectionsMetadata.load_from(None).collections == {}

    filename = tmp_path / "collection-meta.yaml"
    filename.unlink(missing_ok=True)

    assert CollectionsMetadata.load_from(tmp_path).collections == {}

    filename.write_text(
        r"""---
collections: {}
"""
    )

    assert CollectionsMetadata.load_from(tmp_path).collections == {}

    filename.write_text(
        r"""---
collections:
  foo.bar:
    repository: https://github.com/ansible-collections/collection_template
"""
    )

    meta = CollectionsMetadata.load_from(tmp_path)
    assert meta.collections == {
        "foo.bar": CollectionMetadata(
            repository="https://github.com/ansible-collections/collection_template"
        ),
    }
    assert meta.get_meta("foo.bar") == CollectionMetadata(
        repository="https://github.com/ansible-collections/collection_template"
    )

    assert "not.there" not in meta.collections
    assert meta.get_meta("not.there") == CollectionMetadata()
    assert "not.there" in meta.collections


def test_removal_info_methods():
    ri = RemovalInformation(reason="considered-unmaintained", major_version=10)
    assert ri.get_updates_including_indirect() == []
    assert ri.is_deprecated() is True

    ri = RemovalInformation(
        reason="considered-unmaintained",
        major_version=10,
        announce_version="1.2.0",
        updates=[
            RemovalUpdate(cancelled_version="1.3.0"),
            RemovalUpdate(redeprecated_version="1.4.0"),
            RemovalUpdate(removed_version="1.5.0"),
        ],
    )
    assert ri.get_updates_including_indirect() == [
        RemovalUpdate(deprecated_version="1.2.0"),
        RemovalUpdate(cancelled_version="1.3.0"),
        RemovalUpdate(redeprecated_version="1.4.0"),
        RemovalUpdate(removed_version="1.5.0"),
    ]
    assert ri.is_deprecated() is True

    ri = RemovalInformation(
        reason="considered-unmaintained",
        major_version=10,
        announce_version="1.2.0",
        updates=[
            RemovalUpdate(cancelled_version="1.3.0"),
            RemovalUpdate(redeprecated_version="1.4.0"),
            RemovalUpdate(removed_version="1.5.0"),
            RemovalUpdate(readded_version="1.6.0"),
            RemovalUpdate(deprecated_version="1.7.0"),
            RemovalUpdate(cancelled_version="1.8.0"),
        ],
    )
    assert ri.get_updates_including_indirect() == [
        RemovalUpdate(deprecated_version="1.2.0"),
        RemovalUpdate(cancelled_version="1.3.0"),
        RemovalUpdate(redeprecated_version="1.4.0"),
        RemovalUpdate(removed_version="1.5.0"),
        RemovalUpdate(readded_version="1.6.0"),
        RemovalUpdate(deprecated_version="1.7.0"),
        RemovalUpdate(cancelled_version="1.8.0"),
    ]
    assert ri.is_deprecated() is False
